import re

from buidl.hd import HDPublicKey, is_valid_bip32_path
from buidl.helper import is_intable, sha256, uses_only_hex_chars
from buidl.op import number_to_op_code
from buidl.script import P2WSHScriptPubKey, WitnessScript

import json


DESCRIPTOR_INPUT_CHARSET = "0123456789()[],'/*abcdefgh@:$%{}IJKLMNOPQRSTUVWXYZ&+-.;<=>?!^_|~ijklmnopqrstuvwxyzABCDEFGH`#\"\\ "
DESCRIPTOR_CHECKSUM_CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"


def calc_poly_mod(c, val):
    c0 = c >> 35
    c = ((c & 0x7FFFFFFFF) << 5) ^ val
    if c0 & 1:
        c ^= 0xF5DEE51989
    if c0 & 2:
        c ^= 0xA9FDCA3312
    if c0 & 4:
        c ^= 0x1BAB10E32D
    if c0 & 8:
        c ^= 0x3706B1677A
    if c0 & 16:
        c ^= 0x644D626FFD
    return c


def calc_core_checksum(output_descriptor):
    """
    A 40-bit (!) digest that bitcoin core uses for output descriptors
    """

    c = 1
    cls = 0
    clscount = 0
    for ch in output_descriptor:
        pos = DESCRIPTOR_INPUT_CHARSET.find(ch)
        if pos == -1:
            raise ValueError(
                f"Invalid character `{ch}` in output descriptor: {output_descriptor}"
            )
        c = calc_poly_mod(c, pos & 31)
        cls = cls * 3 + (pos >> 5)
        clscount += 1
        if clscount == 3:
            c = calc_poly_mod(c, cls)
            cls = 0
            clscount = 0
    if clscount > 0:
        c = calc_poly_mod(c, cls)
    for j in range(0, 8):
        c = calc_poly_mod(c, 0)
    c ^= 1

    ret = [None] * 8
    for j in range(0, 8):
        ret[j] = DESCRIPTOR_CHECKSUM_CHARSET[(c >> (5 * (7 - j))) & 31]
    return "".join(ret)


def is_valid_xfp_hex(string):
    return len(string) == 8 and uses_only_hex_chars(string)


def parse_full_key_record(key_record_str):
    """
    A full key record will come from your Coordinator and include a reference to an account index.
    It will look something like this:
    [c7d0648a/48h/1h/0h/2h]tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr/0/*

    A full key record is basically a partial key record, with a trailing /{account-index}/*
    """

    # Validate that it appears to be a full key record:
    parts = key_record_str.split("/")
    if parts[-1] != "*":
        raise ValueError(
            "Invalid full key record, does not end with a *: {key_record_str}"
        )
    if not is_intable(parts[-2]):
        raise ValueError(
            "Invalid full key record, account index `{parts[-2]}` is not an int: {key_record_str}"
        )

    # Now we strip off the trailing account index and *, and parse the rest as a partial key record
    partial_key_record_str = "/".join(parts[0 : len(parts) - 2])

    to_return = parse_partial_key_record(key_record_str=partial_key_record_str)
    to_return["account_index"] = int(parts[-2])
    to_return["xpub_parent"] = to_return.pop("xpub")

    try:
        parent_pubkey_obj = HDPublicKey.parse(s=to_return["xpub_parent"])
        to_return["xpub_child"] = parent_pubkey_obj.child(
            index=to_return["account_index"]
        ).xpub()
    except ValueError:
        raise ValueError(
            f"Invalid parent xpub {to_return['xpub_parent']} in key record: {key_record_str}"
        )

    return to_return


def parse_partial_key_record(key_record_str):
    """
    A partial key record will come from your Signer and include no references to change derivation.
    It will look something like this:
    [c7d0648a/48h/1h/0h/2h]tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr
    """

    key_record_re = re.match(
        r"\[([0-9a-f]{8})\*?(.*?)\]([0-9A-Za-z].*)", key_record_str
    )
    if key_record_re is None:
        raise ValueError(f"Invalid key record: {key_record_str}")

    xfp, path, xpub = key_record_re.groups()
    # Note that we don't validate xfp because the regex already tells us it's hex

    path = "m" + path
    if not is_valid_bip32_path(path):
        raise ValueError(f"Invalid BIP32 path {path} in key record: {key_record_str}")

    try:
        pubkey_obj = HDPublicKey.parse(s=xpub)
        network = pubkey_obj.network
    except ValueError:
        raise ValueError(f"Invalid xpub {xpub} in key record: {key_record_str}")

    return {
        "xfp": xfp,
        "path": path,
        "xpub": xpub,
        "network": network,
    }


def parse_any_key_record(key_record_str):
    """
    Try to parse a key record as full key record, and if not parse as a partial key record

    """
    try:
        return parse_full_key_record(key_record_str)
    except ValueError:
        return parse_partial_key_record(key_record_str)


class P2WSHSortedMulti:

    # TODO: make an inheritable base descriptor class that this inherits from

    def __init__(
        self,
        quorum_m,  # m as in m-of-n
        key_records=[],  # pubkeys required to sign
        checksum="",
        sort_key_records=True,
    ):
        if type(quorum_m) is not int or quorum_m < 1:
            raise ValueError(f"quorum_m must be a positive int: {quorum_m}")
        self.quorum_m = quorum_m

        if not key_records:
            raise ValueError("No key_records supplied")

        key_records_to_save, network = [], None
        for key_record in key_records:
            # TODO: does bitcoin core have a standard to enforce for h vs ' in bip32 path?
            path = key_record.get("path")
            if not is_valid_bip32_path(path):
                raise ValueError(
                    f"Invalid BIP32 path `{path}` in key record: {key_record}"
                )

            xfp_hex = key_record.get("xfp")
            if not is_valid_xfp_hex(xfp_hex):
                raise ValueError(
                    f"Invalid hex fingerprint `{xfp_hex}` in key record: {key_record}"
                )

            account_index = key_record.get("account_index")
            if type(account_index) is not int:
                raise ValueError(
                    f"Invalid account index `{account_index}` in key record: {key_record}"
                )

            xpub_parent = key_record.get("xpub_parent")
            try:
                hdpubkey_obj = HDPublicKey.parse(xpub_parent)
                # get rid of slip132 version byte (if it exists) as this will alter the checksum calculation
                hdpubkey_obj_attrs = vars(hdpubkey_obj)
                del hdpubkey_obj_attrs["pub_version"]
                del hdpubkey_obj_attrs["_raw"]
                xpub_to_use = HDPublicKey(**hdpubkey_obj_attrs).xpub()
            except ValueError:
                raise ValueError(
                    f"Invalid xpub_parent `{xpub_parent}` in key record: {key_record}"
                )

            if network is None:
                # This is the first key_record in our loop
                network = hdpubkey_obj.network
            else:
                # Validate that we haven't changed networks
                if hdpubkey_obj.network != network:
                    raise ValueError(
                        f"Network mismatch: network is set to {network} but xpub_parent is {xpub_parent}"
                    )

            key_records_to_save.append(
                {
                    "path": path,
                    "xfp": xfp_hex,
                    "xpub_parent": xpub_to_use,
                    "account_index": account_index,
                }
            )

        if sort_key_records:
            # Sort lexicographically based on parent xpub
            key_records_to_save = sorted(
                key_records_to_save, key=lambda k: k["xpub_parent"]
            )

        # Generate descriptor text (not part of above loop due to sort_key_records)
        descriptor_text = f"wsh(sortedmulti({quorum_m}"
        for kr in key_records_to_save:
            descriptor_text += f",[{kr['xfp']}{kr['path'][1:]}]{kr['xpub_parent']}/{kr['account_index']}/*"
        descriptor_text += "))"
        self.descriptor_text = descriptor_text
        self.key_records = key_records_to_save
        self.network = network

        calculated_checksum = calc_core_checksum(descriptor_text)

        if checksum:
            # test that it matches
            if calculated_checksum != checksum:
                raise ValueError(
                    f"Calculated checksum `{calculated_checksum}` != supplied checksum `{checksum}`"
                )
        self.checksum = calculated_checksum

    def __repr__(self):
        return self.descriptor_text + "#" + self.checksum

    @property
    def quorum_n(self):
        return len(self.key_records)

    @property
    def m_of_n(self):
        return f"{self.quorum_m}-of-{self.quorum_n}"

    @classmethod
    def parse(cls, output_record):
        # Fix strange slashes that some software (Specter-Desktop) may export
        output_record = output_record.strip().replace(r"\/", "/")

        # Regex match the string
        re_output_results = re.match(
            r".*wsh\(sortedmulti\(([0-9]*),(.*)\)\)(\#[qpzry9x8gf2tvdw0s3jn54khce6mua7l]{8})?.*",
            output_record,
        )
        if re_output_results is None:
            raise ValueError(f"Not a valid wsh sortedmulti: {output_record}")

        quorum_m_str, key_records_str, checksum = re_output_results.groups()

        if "#" in output_record:
            if checksum:
                # get rid of leading # from capture group
                # TODO: is there a more elegant way to do this?
                checksum = checksum[1:]
            else:
                err_msg = f"Could not parse checksum in output_record: {output_record}"
                err_msg += "\n\nPerhaps try again this with no checksum (no #foo at the end of your output_record)?"
                raise ValueError(err_msg)

        quorum_m_int = int(quorum_m_str)

        key_records = []
        for key_record_str in key_records_str.split(","):
            # A full key record will look something like this:
            # [c7d0648a/48h/1h/0h/2h]tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr/0/*'
            key_records.append(parse_full_key_record(key_record_str))

        if quorum_m_int > len(key_records):
            raise ValueError(
                f"Malformed threshold {quorum_m_int}-of-{len(key_records)} (m must be less than n) in {output_record}"
            )

        return cls(
            quorum_m=quorum_m_int,
            key_records=key_records,
            sort_key_records=False,
            checksum=checksum,
        )

    def get_address(self, offset=0, is_change=False, sort_keys=True):
        """
        If is_change=True, then we display change addresses.
        If is_change=False we display receive addresses.

        sort_keys is for expert users only and should be left as True
        """
        assert type(is_change) is bool, is_change
        assert type(offset) is int and offset >= 0, offset

        sec_hexes_to_use = []
        for key_record in self.key_records:
            hdpubkey = HDPublicKey.parse(key_record["xpub_parent"])
            if is_change is True:
                account = key_record["account_index"] + 1
            else:
                account = key_record["account_index"]
            leaf_xpub = hdpubkey.child(account).child(offset)
            sec_hexes_to_use.append(leaf_xpub.sec().hex())

        commands = [number_to_op_code(self.quorum_m)]
        if sort_keys:
            # BIP67 lexicographical sorting for sortedmulti
            commands.extend([bytes.fromhex(x) for x in sorted(sec_hexes_to_use)])
        else:
            commands.extend([bytes.fromhex(x) for x in sec_hexes_to_use])

        commands.append(number_to_op_code(len(self.key_records)))
        commands.append(174)  # OP_CHECKMULTISIG

        witness_script = WitnessScript(commands)
        redeem_script = P2WSHScriptPubKey(sha256(witness_script.raw_serialize()))
        return redeem_script.address(network=self.network)

    def caravan_export(self, wallet_name="p2wsh", key_record_names=[]):
        if key_record_names and len(key_record_names) != len(self.key_records):
            raise ValueError(
                f"{len(self.key_records)} key records but only {len(key_record_names)} names supplied: {key_record_names}"
            )

        to_return = {
            "name": wallet_name,
            "addressType": "P2WSH",
            "network": self.network,
            "client": {"type": "public"},  # node connection instructions
            "quorum": {
                "requiredSigners": self.quorum_m,
                "totalSigners": len(self.key_records),
            },
            "extendedPublicKeys": [],
            "startingAddressIndex": self.key_records[0]["account_index"],
        }
        for cnt, key_record in enumerate(self.key_records):
            to_append = {
                "bip32Path": key_record["path"].lower().replace("h", "'"),
                "xpub": key_record["xpub_parent"],
                "xfp": key_record["xfp"],
            }
            if key_record_names:
                name = key_record_names[cnt]
            else:
                # Generic/deterministic name: "Seed A"
                seed_letter = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[cnt]
                name = f"Seed {seed_letter}"
            to_append["name"] = name
            to_return["extendedPublicKeys"].append(to_append)

        return json.dumps(to_return)



from unittest import TestCase

from buidl.pecc import FieldElement, Point


class FieldElementTest(TestCase):
    def test_ne(self):
        a = FieldElement(2, 31)
        b = FieldElement(2, 31)
        c = FieldElement(15, 31)
        self.assertEqual(a, b)
        self.assertTrue(a != c)
        self.assertFalse(a != b)

    def test_add(self):
        a = FieldElement(2, 31)
        b = FieldElement(15, 31)
        self.assertEqual(a + b, FieldElement(17, 31))
        a = FieldElement(17, 31)
        b = FieldElement(21, 31)
        self.assertEqual(a + b, FieldElement(7, 31))

    def test_sub(self):
        a = FieldElement(29, 31)
        b = FieldElement(4, 31)
        self.assertEqual(a - b, FieldElement(25, 31))
        a = FieldElement(15, 31)
        b = FieldElement(30, 31)
        self.assertEqual(a - b, FieldElement(16, 31))

    def test_mul(self):
        a = FieldElement(24, 31)
        b = FieldElement(19, 31)
        self.assertEqual(a * b, FieldElement(22, 31))

    def test_pow(self):
        a = FieldElement(17, 31)
        self.assertEqual(a ** 3, FieldElement(15, 31))
        a = FieldElement(5, 31)
        b = FieldElement(18, 31)
        self.assertEqual(a ** 5 * b, FieldElement(16, 31))

    def test_div(self):
        a = FieldElement(3, 31)
        b = FieldElement(24, 31)
        self.assertEqual(a / b, FieldElement(4, 31))
        a = FieldElement(17, 31)
        self.assertEqual(a ** -3, FieldElement(29, 31))
        a = FieldElement(4, 31)
        b = FieldElement(11, 31)
        self.assertEqual(a ** -4 * b, FieldElement(13, 31))


class PointTest(TestCase):
    def test_ne(self):
        a = Point(x=3, y=-7, a=5, b=7)
        b = Point(x=18, y=77, a=5, b=7)
        self.assertTrue(a != b)
        self.assertFalse(a != a)

    def test_on_curve(self):
        with self.assertRaises(ValueError):
            Point(x=-2, y=4, a=5, b=7)
        # these should not raise an error
        Point(x=3, y=-7, a=5, b=7)
        Point(x=18, y=77, a=5, b=7)

    def test_add0(self):
        a = Point(x=None, y=None, a=5, b=7)
        b = Point(x=2, y=5, a=5, b=7)
        c = Point(x=2, y=-5, a=5, b=7)
        self.assertEqual(a + b, b)
        self.assertEqual(b + a, b)
        self.assertEqual(b + c, a)

    def test_add1(self):
        a = Point(x=3, y=7, a=5, b=7)
        b = Point(x=-1, y=-1, a=5, b=7)
        self.assertEqual(a + b, Point(x=2, y=-5, a=5, b=7))

    def test_add2(self):
        a = Point(x=-1, y=1, a=5, b=7)
        self.assertEqual(a + a, Point(x=18, y=-77, a=5, b=7))


class ECCTest(TestCase):
    def test_on_curve(self):
        # tests the following points whether they are on the curve or not
        # on curve y^2=x^3-7 over F_223:
        # (192,105) (17,56) (200,119) (1,193) (42,99)
        # the ones that aren't should raise a ValueError
        prime = 223
        a = FieldElement(0, prime)
        b = FieldElement(7, prime)

        valid_points = ((192, 105), (17, 56), (1, 193))
        invalid_points = ((200, 119), (42, 99))

        # iterate over valid points
        for x_raw, y_raw in valid_points:
            # Initialize points this way:
            # x = FieldElement(x_raw, prime)
            # y = FieldElement(y_raw, prime)
            # Point(x, y, a, b)
            x = FieldElement(x_raw, prime)
            y = FieldElement(y_raw, prime)
            # Creating the point should not result in an error
            Point(x, y, a, b)

        # iterate over invalid points
        for x_raw, y_raw in invalid_points:
            # Initialize points this way:
            # x = FieldElement(x_raw, prime)
            # y = FieldElement(y_raw, prime)
            # Point(x, y, a, b)
            x = FieldElement(x_raw, prime)
            y = FieldElement(y_raw, prime)
            # check that creating the point results in a ValueError
            # with self.assertRaises(ValueError):
            #     Point(x, y, a, b)
            with self.assertRaises(ValueError):
                Point(x, y, a, b)

    def test_add(self):
        # tests the following additions on curve y^2=x^3-7 over F_223:
        # (192,105) + (17,56)
        # (47,71) + (117,141)
        # (143,98) + (76,66)
        prime = 223
        a = FieldElement(0, prime)
        b = FieldElement(7, prime)

        additions = (
            # (x1, y1, x2, y2, x3, y3)
            (192, 105, 17, 56, 170, 142),
            (47, 71, 117, 141, 60, 139),
            (143, 98, 76, 66, 47, 71),
        )
        # iterate over the additions
        for x1_raw, y1_raw, x2_raw, y2_raw, x3_raw, y3_raw in additions:
            # Initialize points this way:
            # x1 = FieldElement(x1_raw, prime)
            # y1 = FieldElement(y1_raw, prime)
            # p1 = Point(x1, y1, a, b)
            # x2 = FieldElement(x2_raw, prime)
            # y2 = FieldElement(y2_raw, prime)
            # p2 = Point(x2, y2, a, b)
            # x3 = FieldElement(x3_raw, prime)
            # y3 = FieldElement(y3_raw, prime)
            # p3 = Point(x3, y3, a, b)
            x1 = FieldElement(x1_raw, prime)
            y1 = FieldElement(y1_raw, prime)
            p1 = Point(x1, y1, a, b)
            x2 = FieldElement(x2_raw, prime)
            y2 = FieldElement(y2_raw, prime)
            p2 = Point(x2, y2, a, b)
            x3 = FieldElement(x3_raw, prime)
            y3 = FieldElement(y3_raw, prime)
            p3 = Point(x3, y3, a, b)
            # check that p1 + p2 == p3
            self.assertEqual(p1 + p2, p3)

    def test_rmul(self):
        # tests the following scalar multiplications
        # 2*(192,105)
        # 2*(143,98)
        # 2*(47,71)
        # 4*(47,71)
        # 8*(47,71)
        # 21*(47,71)
        prime = 223
        a = FieldElement(0, prime)
        b = FieldElement(7, prime)

        multiplications = (
            # (coefficient, x1, y1, x2, y2)
            (2, 192, 105, 49, 71),
            (2, 143, 98, 64, 168),
            (2, 47, 71, 36, 111),
            (4, 47, 71, 194, 51),
            (8, 47, 71, 116, 55),
            (21, 47, 71, None, None),
        )

        # iterate over the multiplications
        for s, x1_raw, y1_raw, x2_raw, y2_raw in multiplications:
            # Initialize points this way:
            # x1 = FieldElement(x1_raw, prime)
            # y1 = FieldElement(y1_raw, prime)
            # p1 = Point(x1, y1, a, b)
            x1 = FieldElement(x1_raw, prime)
            y1 = FieldElement(y1_raw, prime)
            p1 = Point(x1, y1, a, b)
            # initialize the second point based on whether it's the point at infinity
            # x2 = FieldElement(x2_raw, prime)
            # y2 = FieldElement(y2_raw, prime)
            # p2 = Point(x2, y2, a, b)
            if x2_raw is None:
                p2 = Point(None, None, a, b)
            else:
                x2 = FieldElement(x2_raw, prime)
                y2 = FieldElement(y2_raw, prime)
                p2 = Point(x2, y2, a, b)
            # check that the product is equal to the expected point
            self.assertEqual(s * p1, p2)



from unittest import TestCase
from io import BytesIO

from buidl.block import Block


class BlockTest(TestCase):
    def test_parse(self):
        block_raw = bytes.fromhex(
            "0100000000000000000000000000000000000000000000000000000000000000000000003ba3edfd7a7b12b27ac72c3e67768f617fc81bc3888a51323a9fb8aa4b1e5e4a29ab5f49ffff001d1dac2b7c0101000000010000000000000000000000000000000000000000000000000000000000000000ffffffff4d04ffff001d0104455468652054696d65732030332f4a616e2f32303039204368616e63656c6c6f72206f6e206272696e6b206f66207365636f6e64206261696c6f757420666f722062616e6b73ffffffff0100f2052a01000000434104678afdb0fe5548271967f1a67130b7105cd6a828e03909a67962e0ea1f61deb649f6bc3f4cef38c4f35504e51ec112de5c384df7ba0b8d578a4c702b6bf11d5fac00000000"
        )
        stream = BytesIO(block_raw)
        block = Block.parse(stream)
        self.assertEqual(block.merkle_root.hex(), block.tx_hashes[0].hex())

    def test_parse_header(self):
        block_raw = bytes.fromhex(
            "020000208ec39428b17323fa0ddec8e887b4a7c53b8c0a0a220cfd0000000000000000005b0750fce0a889502d40508d39576821155e9c9e3f5c3157f961db38fd8b25be1e77a759e93c0118a4ffd71d"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertEqual(block.version, 0x20000002)
        want = bytes.fromhex(
            "000000000000000000fd0c220a0a8c3bc5a7b487e8c8de0dfa2373b12894c38e"
        )
        self.assertEqual(block.prev_block, want)
        want = bytes.fromhex(
            "be258bfd38db61f957315c3f9e9c5e15216857398d50402d5089a8e0fc50075b"
        )
        self.assertEqual(block.merkle_root, want)
        self.assertEqual(block.timestamp, 0x59A7771E)
        self.assertEqual(block.bits, bytes.fromhex("e93c0118"))
        self.assertEqual(block.nonce, bytes.fromhex("a4ffd71d"))

    def test_serialize(self):
        block_raw = bytes.fromhex(
            "020000208ec39428b17323fa0ddec8e887b4a7c53b8c0a0a220cfd0000000000000000005b0750fce0a889502d40508d39576821155e9c9e3f5c3157f961db38fd8b25be1e77a759e93c0118a4ffd71d"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertEqual(block.serialize(), block_raw)

    def test_hash(self):
        block_raw = bytes.fromhex(
            "020000208ec39428b17323fa0ddec8e887b4a7c53b8c0a0a220cfd0000000000000000005b0750fce0a889502d40508d39576821155e9c9e3f5c3157f961db38fd8b25be1e77a759e93c0118a4ffd71d"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertEqual(
            block.hash(),
            bytes.fromhex(
                "0000000000000000007e9e4c586439b0cdbe13b1370bdd9435d76a644d047523"
            ),
        )

    def test_bip9(self):
        block_raw = bytes.fromhex(
            "020000208ec39428b17323fa0ddec8e887b4a7c53b8c0a0a220cfd0000000000000000005b0750fce0a889502d40508d39576821155e9c9e3f5c3157f961db38fd8b25be1e77a759e93c0118a4ffd71d"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertTrue(block.bip9())
        block_raw = bytes.fromhex(
            "0400000039fa821848781f027a2e6dfabbf6bda920d9ae61b63400030000000000000000ecae536a304042e3154be0e3e9a8220e5568c3433a9ab49ac4cbb74f8df8e8b0cc2acf569fb9061806652c27"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertFalse(block.bip9())

    def test_bip91(self):
        block_raw = bytes.fromhex(
            "1200002028856ec5bca29cf76980d368b0a163a0bb81fc192951270100000000000000003288f32a2831833c31a25401c52093eb545d28157e200a64b21b3ae8f21c507401877b5935470118144dbfd1"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertTrue(block.bip91())
        block_raw = bytes.fromhex(
            "020000208ec39428b17323fa0ddec8e887b4a7c53b8c0a0a220cfd0000000000000000005b0750fce0a889502d40508d39576821155e9c9e3f5c3157f961db38fd8b25be1e77a759e93c0118a4ffd71d"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertFalse(block.bip91())

    def test_bip141(self):
        block_raw = bytes.fromhex(
            "020000208ec39428b17323fa0ddec8e887b4a7c53b8c0a0a220cfd0000000000000000005b0750fce0a889502d40508d39576821155e9c9e3f5c3157f961db38fd8b25be1e77a759e93c0118a4ffd71d"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertTrue(block.bip141())
        block_raw = bytes.fromhex(
            "0000002066f09203c1cf5ef1531f24ed21b1915ae9abeb691f0d2e0100000000000000003de0976428ce56125351bae62c5b8b8c79d8297c702ea05d60feabb4ed188b59c36fa759e93c0118b74b2618"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertFalse(block.bip141())

    def test_target(self):
        block_raw = bytes.fromhex(
            "020000208ec39428b17323fa0ddec8e887b4a7c53b8c0a0a220cfd0000000000000000005b0750fce0a889502d40508d39576821155e9c9e3f5c3157f961db38fd8b25be1e77a759e93c0118a4ffd71d"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertEqual(
            block.target(), 0x13CE9000000000000000000000000000000000000000000
        )
        self.assertEqual(int(block.difficulty()), 888171856257)

    def test_check_pow(self):
        block_raw = bytes.fromhex(
            "04000000fbedbbf0cfdaf278c094f187f2eb987c86a199da22bbb20400000000000000007b7697b29129648fa08b4bcd13c9d5e60abb973a1efac9c8d573c71c807c56c3d6213557faa80518c3737ec1"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertTrue(block.check_pow())
        block_raw = bytes.fromhex(
            "04000000fbedbbf0cfdaf278c094f187f2eb987c86a199da22bbb20400000000000000007b7697b29129648fa08b4bcd13c9d5e60abb973a1efac9c8d573c71c807c56c3d6213557faa80518c3737ec0"
        )
        stream = BytesIO(block_raw)
        block = Block.parse_header(stream)
        self.assertFalse(block.check_pow())

    def test_validate_merkle_root(self):
        hashes_hex = [
            "f54cb69e5dc1bd38ee6901e4ec2007a5030e14bdd60afb4d2f3428c88eea17c1",
            "c57c2d678da0a7ee8cfa058f1cf49bfcb00ae21eda966640e312b464414731c1",
            "b027077c94668a84a5d0e72ac0020bae3838cb7f9ee3fa4e81d1eecf6eda91f3",
            "8131a1b8ec3a815b4800b43dff6c6963c75193c4190ec946b93245a9928a233d",
            "ae7d63ffcb3ae2bc0681eca0df10dda3ca36dedb9dbf49e33c5fbe33262f0910",
            "61a14b1bbdcdda8a22e61036839e8b110913832efd4b086948a6a64fd5b3377d",
            "fc7051c8b536ac87344c5497595d5d2ffdaba471c73fae15fe9228547ea71881",
            "77386a46e26f69b3cd435aa4faac932027f58d0b7252e62fb6c9c2489887f6df",
            "59cbc055ccd26a2c4c4df2770382c7fea135c56d9e75d3f758ac465f74c025b8",
            "7c2bf5687f19785a61be9f46e031ba041c7f93e2b7e9212799d84ba052395195",
            "08598eebd94c18b0d59ac921e9ba99e2b8ab7d9fccde7d44f2bd4d5e2e726d2e",
            "f0bb99ef46b029dd6f714e4b12a7d796258c48fee57324ebdc0bbc4700753ab1",
        ]
        hashes = [bytes.fromhex(x) for x in hashes_hex]
        stream = BytesIO(
            bytes.fromhex(
                "00000020fcb19f7895db08cadc9573e7915e3919fb76d59868a51d995201000000000000acbcab8bcc1af95d8d563b77d24c3d19b18f1486383d75a5085c4e86c86beed691cfa85916ca061a00000000"
            )
        )
        block = Block.parse_header(stream)
        block.tx_hashes = hashes
        self.assertTrue(block.validate_merkle_root())



from io import BytesIO

from buidl.bech32 import decode_bech32, encode_bech32_checksum
from buidl.helper import (
    decode_base58,
    encode_base58_checksum,
    encode_varstr,
    hash160,
    little_endian_to_int,
    int_to_byte,
    int_to_little_endian,
    read_varint,
    sha256,
)
from buidl.op import (
    number_to_op_code,
    op_code_to_number,
    op_equal,
    op_hash160,
    op_verify,
    OP_CODE_FUNCTIONS,
    OP_CODE_NAMES,
)


class Script:
    def __init__(self, commands=None, coinbase=None):
        if commands is None:
            self.commands = []
        else:
            self.commands = commands
        self.coinbase = coinbase

    def __repr__(self):
        result = ""
        for command in self.commands:
            if type(command) == int:
                if OP_CODE_NAMES.get(command):
                    name = OP_CODE_NAMES.get(command)
                else:
                    name = "OP_[{}]".format(command)
                result += "{} ".format(name)
            else:
                result += "{} ".format(command.hex())
        return result

    def __eq__(self, other):
        return self.commands == other.commands

    def __add__(self, other):
        return Script(self.commands + other.commands)

    @classmethod
    def parse(cls, s, coinbase_mode=False):
        # get the length of the entire field
        length = read_varint(s)
        if coinbase_mode:
            return cls([], coinbase=s.read(length))
        # initialize the commands array
        commands = []
        # initialize the number of bytes we've read to 0
        count = 0
        # loop until we've read length bytes
        while count < length:
            # get the current byte
            current = s.read(1)
            # increment the bytes we've read
            count += 1
            # convert the current byte to an integer
            current_byte = current[0]
            # if the current byte is between 1 and 75 inclusive
            if current_byte >= 1 and current_byte <= 75:
                # we have an command set n to be the current byte
                n = current_byte
                # add the next n bytes as an command
                commands.append(s.read(n))
                # increase the count by n
                count += n
            elif current_byte == 76:
                # op_pushdata1
                data_length = little_endian_to_int(s.read(1))
                commands.append(s.read(data_length))
                count += data_length + 1
            elif current_byte == 77:
                # op_pushdata2
                data_length = little_endian_to_int(s.read(2))
                commands.append(s.read(data_length))
                count += data_length + 2
            elif current_byte == 78:
                # op_pushdata4
                data_length = little_endian_to_int(s.read(4))
                commands.append(s.read(data_length))
                count += data_length + 4
            else:
                # we have an op code. set the current byte to op_code
                op_code = current_byte
                # add the op_code to the list of commands
                commands.append(op_code)
        if count != length:
            raise RuntimeError("parsing script failed")
        return cls(commands)

    def raw_serialize(self):
        if self.coinbase:
            return self.coinbase
        # initialize what we'll send back
        result = b""
        # go through each command
        for command in self.commands:
            # if the command is an integer, it's an op code
            if type(command) == int:
                # turn the command into a single byte integer using int_to_byte
                result += int_to_byte(command)
            else:
                # otherwise, this is an element
                # get the length in bytes
                length = len(command)
                # for large lengths, we have to use a pushdata op code
                if length < 75:
                    # turn the length into a single byte integer
                    result += int_to_byte(length)
                elif length > 75 and length < 0x100:
                    # 76 is pushdata1
                    result += int_to_byte(76)
                    result += int_to_byte(length)
                elif length >= 0x100 and length <= 520:
                    # 77 is pushdata2
                    result += int_to_byte(77)
                    result += int_to_little_endian(length, 2)
                else:
                    raise ValueError("too long a command")
                result += command
        return result

    def serialize(self):
        # get the raw serialization (no prepended length)
        result = self.raw_serialize()
        # encode_varstr the result
        return encode_varstr(result)

    def evaluate(self, z, witness):
        # create a copy as we may need to add to this list if we have a
        # RedeemScript
        commands = self.commands[:]
        stack = []
        altstack = []
        while len(commands) > 0:
            command = commands.pop(0)
            if type(command) == int:
                # do what the op code says
                operation = OP_CODE_FUNCTIONS[command]
                if command in (99, 100):
                    # op_if/op_notif require the commands array
                    if not operation(stack, commands):
                        print("bad op: {}".format(OP_CODE_NAMES[command]))
                        return False
                elif command in (107, 108):
                    # op_toaltstack/op_fromaltstack require the altstack
                    if not operation(stack, altstack):
                        print("bad op: {}".format(OP_CODE_NAMES[command]))
                        return False
                elif command in (172, 173, 174, 175):
                    # these are signing operations, they need a sig_hash
                    # to check against
                    if not operation(stack, z):
                        print("bad op: {}".format(OP_CODE_NAMES[command]))
                        return False
                else:
                    if not operation(stack):
                        print("bad op: {}".format(OP_CODE_NAMES[command]))
                        return False
            else:
                # add the command to the stack
                stack.append(command)
                # p2sh rule. if the next three commands are:
                # OP_HASH160 <20 byte hash> OP_EQUAL this is the RedeemScript
                # OP_HASH160 == 0xa9 and OP_EQUAL == 0x87
                if (
                    len(commands) == 3
                    and commands[0] == 0xA9
                    and type(commands[1]) == bytes
                    and len(commands[1]) == 20
                    and commands[2] == 0x87
                ):
                    redeem_script = encode_varstr(command)
                    # we execute the next three op codes
                    commands.pop()
                    h160 = commands.pop()
                    commands.pop()
                    if not op_hash160(stack):
                        return False
                    stack.append(h160)
                    if not op_equal(stack):
                        return False
                    # final result should be a 1
                    if not op_verify(stack):
                        print("bad p2sh h160")
                        return False
                    # hashes match! now add the RedeemScript
                    stream = BytesIO(redeem_script)
                    commands.extend(Script.parse(stream).commands)
                # witness program version 0 rule. if stack commands are:
                # 0 <20 byte hash> this is p2wpkh
                if len(stack) == 2 and stack[0] == b"" and len(stack[1]) == 20:
                    h160 = stack.pop()
                    stack.pop()
                    commands.extend(witness.items)
                    commands.extend(P2PKHScriptPubKey(h160).commands)
                # witness program version 0 rule. if stack commands are:
                # 0 <32 byte hash> this is p2wsh
                if len(stack) == 2 and stack[0] == b"" and len(stack[1]) == 32:
                    s256 = stack.pop()
                    stack.pop()
                    commands.extend(witness.items[:-1])
                    witness_script = witness.items[-1]
                    if s256 != sha256(witness_script):
                        print(
                            "bad sha256 {} vs {}".format(
                                s256.hex(), sha256(witness_script).hex()
                            )
                        )
                        return False
                    # hashes match! now add the Witness Script
                    stream = BytesIO(encode_varstr(witness_script))
                    witness_script_commands = Script.parse(stream).commands
                    commands.extend(witness_script_commands)
        if len(stack) == 0:
            return False
        if stack.pop() == b"":
            return False
        return True

    def is_p2pkh(self):
        """Returns whether the script follows the
        OP_DUP OP_HASH160 <20 byte hash> OP_EQUALVERIFY OP_CHECKSIG pattern."""
        # there should be exactly 5 commands
        # OP_DUP (0x76), OP_HASH160 (0xa9), 20-byte hash, OP_EQUALVERIFY (0x88),
        # OP_CHECKSIG (0xac)
        return (
            len(self.commands) == 5
            and self.commands[0] == 0x76
            and self.commands[1] == 0xA9
            and type(self.commands[2]) == bytes
            and len(self.commands[2]) == 20
            and self.commands[3] == 0x88
            and self.commands[4] == 0xAC
        )

    def is_p2sh(self):
        """Returns whether the script follows the
        OP_HASH160 <20 byte hash> OP_EQUAL pattern."""
        # there should be exactly 3 commands
        # OP_HASH160 (0xa9), 20-byte hash, OP_EQUAL (0x87)
        return (
            len(self.commands) == 3
            and self.commands[0] == 0xA9
            and type(self.commands[1]) == bytes
            and len(self.commands[1]) == 20
            and self.commands[2] == 0x87
        )

    def is_p2wpkh(self):
        """Returns whether the script follows the
        OP_0 <20 byte hash> pattern."""
        return (
            len(self.commands) == 2
            and self.commands[0] == 0x00
            and type(self.commands[1]) == bytes
            and len(self.commands[1]) == 20
        )

    def is_p2wsh(self):
        """Returns whether the script follows the
        OP_0 <32 byte hash> pattern."""
        return (
            len(self.commands) == 2
            and self.commands[0] == 0x00
            and type(self.commands[1]) == bytes
            and len(self.commands[1]) == 32
        )

    def has_op_return(self):
        return 106 in self.commands

    def is_witness_script(self):
        if self.is_p2wpkh() or self.is_p2wsh():
            return True
        else:
            return False


class ScriptPubKey(Script):
    """Represents a ScriptPubKey in a transaction"""

    @classmethod
    def parse(cls, s):
        script_pubkey = super().parse(s)
        if script_pubkey.is_p2pkh():
            return P2PKHScriptPubKey(script_pubkey.commands[2])
        elif script_pubkey.is_p2sh():
            return P2SHScriptPubKey(script_pubkey.commands[1])
        elif script_pubkey.is_p2wpkh():
            return P2WPKHScriptPubKey(script_pubkey.commands[1])
        elif script_pubkey.is_p2wsh():
            return P2WSHScriptPubKey(script_pubkey.commands[1])
        else:
            return script_pubkey

    def redeem_script(self):
        """Convert this ScriptPubKey to its RedeemScript equivalent"""
        return RedeemScript(self.commands)


class P2PKHScriptPubKey(ScriptPubKey):
    def __init__(self, h160):
        super().__init__()
        if type(h160) != bytes:
            raise TypeError("To initialize P2PKHScriptPubKey, a hash160 is needed")
        self.commands = [0x76, 0xA9, h160, 0x88, 0xAC]

    def hash160(self):
        return self.commands[2]

    def address(self, network="mainnet"):
        if network == "mainnet":
            prefix = b"\x00"
        else:
            prefix = b"\x6f"
        # return the encode_base58_checksum the prefix and h160
        return encode_base58_checksum(prefix + self.hash160())


class P2SHScriptPubKey(ScriptPubKey):
    def __init__(self, h160):
        super().__init__()
        if type(h160) != bytes:
            raise TypeError("To initialize P2SHScriptPubKey, a hash160 is needed")
        self.commands = [0xA9, h160, 0x87]

    def hash160(self):
        return self.commands[1]

    def address(self, network="mainnet"):
        if network == "mainnet":
            prefix = b"\x05"
        else:
            prefix = b"\xc4"
        # return the encode_base58_checksum the prefix and h160
        return encode_base58_checksum(prefix + self.hash160())


class RedeemScript(Script):
    """Subclass that represents a RedeemScript for p2sh"""

    def is_p2sh_multisig(self):
        return self.commands[-1] == 174

    def hash160(self):
        """Returns the hash160 of the serialization of the RedeemScript"""
        return hash160(self.raw_serialize())

    def script_pubkey(self):
        """Returns the ScriptPubKey that this RedeemScript corresponds to"""
        return P2SHScriptPubKey(self.hash160())

    def address(self, network="mainnet"):
        """Returns the p2sh address for this RedeemScript"""
        return self.script_pubkey().address(network)

    @classmethod
    def convert(cls, raw_redeem_script):
        stream = BytesIO(encode_varstr(raw_redeem_script))
        return cls.parse(stream)

    @classmethod
    def create_p2sh_multisig(
        cls,
        quorum_m,
        pubkey_hexes,
        sort_keys=True,
        expected_addr=None,
        expected_addr_network="mainnet",
    ):
        """
        Create a p2sh RedeemScript using a configure threshold (quorum_m) of child public keys (in hex).

        To use a custom order of pubkeys, feed them in order and set sort_keys=False

        For safety, you may pass in an expected_addr and the method will throw an error if the derived address doesn't match the expected one.
        """
        # safety checks
        if type(quorum_m) is not int:
            raise ValueError(f"quorum_m must be of type int: {quorum_m}")
        if quorum_m < 1 or quorum_m > len(pubkey_hexes):
            raise ValueError(f"Invalid m-of-n: {quorum_m}-of-{len(pubkey_hexes)}")

        commands = [number_to_op_code(quorum_m)]

        if sort_keys:
            pubkey_hexes = sorted(pubkey_hexes)

        for pubkey_hex in pubkey_hexes:
            # we want these in binary (not hex)
            commands.append(bytes.fromhex(pubkey_hex))

        quorum_n = len(pubkey_hexes)
        commands.append(number_to_op_code(quorum_n))
        commands.append(174)  # OP_CHECKMULTISIG

        to_return = cls(commands)

        if expected_addr:
            calculated_addr = to_return.address(network=expected_addr_network)
            if expected_addr != calculated_addr:
                raise ValueError(
                    f"Expected address {expected_addr} but calculated {calculated_addr}"
                )

        return to_return

    def get_quorum(self):
        """
        Return the m-of-n of this multisig, as in 2-of-3 or 3-of-5
        """
        if not self.is_p2sh_multisig():
            raise ValueError(f"Not p2sh multisig: {self}")
        quorum_m = op_code_to_number(self.commands[0])
        # 3 because quorum_m, OP_CHECKMULTISIG, and bitcoin off-by-one error
        quorum_n = len(self.commands) - 3
        return quorum_m, quorum_n

    def signing_pubkeys(self):
        """
        The pubkeys needed to sign this transaction, typically children derived from xpubs
        """
        if not self.is_p2sh_multisig():
            raise ValueError(f"Not p2sh multisig: {self}")
        return self.commands[1:-2]


class SegwitPubKey(ScriptPubKey):
    def address(self, network="mainnet"):
        """return the bech32 address for the p2wpkh"""
        # witness program is the raw serialization
        witness_program = self.raw_serialize()
        # convert to bech32 address using encode_bech32_checksum
        return encode_bech32_checksum(witness_program, network)

    def p2sh_address(self, network="mainnet"):
        # get the RedeemScript equivalent and get its address
        return self.redeem_script().address(network)


class P2WPKHScriptPubKey(SegwitPubKey):
    def __init__(self, h160):
        super().__init__()
        if type(h160) != bytes:
            raise TypeError("To initialize P2WPKHScriptPubKey, a hash160 is needed")
        self.commands = [0x00, h160]


class P2WSHScriptPubKey(SegwitPubKey):
    def __init__(self, s256):
        super().__init__()
        if type(s256) != bytes:
            raise TypeError("To initialize P2WSHScriptPubKey, a sha256 is needed")
        self.commands = [0x00, s256]


class WitnessScript(Script):
    """Subclass that represents a WitnessScript for p2wsh"""

    @classmethod
    def convert(cls, raw_witness_script):
        stream = BytesIO(encode_varstr(raw_witness_script))
        return cls.parse(stream)

    def sha256(self):
        """Returns the sha256 of the raw serialization for witness program"""
        return sha256(self.raw_serialize())

    def script_pubkey(self):
        """Generates the ScriptPubKey for p2wsh"""
        # get the sha256 of the current script
        s256 = self.sha256()
        # return new p2wsh script using p2wsh_script
        return P2WSHScriptPubKey(s256)

    def address(self, network="mainnet"):
        """Generates a p2wsh address"""
        # grab the entire witness program
        witness_program = self.script_pubkey().raw_serialize()
        # convert to bech32 address using encode_bech32_checksum
        return encode_bech32_checksum(witness_program, network)

    def p2sh_address(self, network="mainnet"):
        """Generates a p2sh-p2wsh address"""
        # the RedeemScript is the p2wsh ScriptPubKey
        redeem_script = self.script_pubkey().redeem_script()
        # return the p2sh address of the RedeemScript (remember network)
        return redeem_script.address(network)

    def is_p2wsh_multisig(self):
        return (
            OP_CODE_NAMES[self.commands[-1]] == "OP_CHECKMULTISIG"
            and type(self.commands[0]) == int
            and type(self.commands[-2]) == int
        )

    def get_quorum(self):
        """
        Return the m-of-n of this multisig, as in 2-of-3 or 3-of-5
        """

        if not self.is_p2wsh_multisig():
            raise ValueError(f"Not a multisig witness script: {self}")

        quorum_m = OP_CODE_NAMES[self.commands[0]].split("OP_")[1]
        quorum_n = OP_CODE_NAMES[self.commands[-2]].split("OP_")[1]

        return int(quorum_m), int(quorum_n)


def address_to_script_pubkey(s):
    if s[:1] in ("1", "m", "n"):
        # p2pkh
        h160 = decode_base58(s)
        return P2PKHScriptPubKey(h160)
    elif s[:1] in ("2", "3"):
        # p2sh
        h160 = decode_base58(s)
        return P2SHScriptPubKey(h160)
    elif s[:3] in ("bc1", "tb1"):
        if len(s) == 42:
            # p2wpkh
            return P2WPKHScriptPubKey(decode_bech32(s)[2])
        elif len(s) == 62:
            # p2wskh
            return P2WSHScriptPubKey(decode_bech32(s)[2])

    raise RuntimeError("unknown type of address: {}".format(s))



#!/usr/bin/env python3
# coding: utf-8

import readline
import sys
from cmd import Cmd
from getpass import getpass
from itertools import combinations
from os import environ
from platform import platform
from pkg_resources import DistributionNotFound, get_distribution

import buidl  # noqa: F401 (used below with pkg_resources for versioning)
from buidl.blinding import blind_xpub, secure_secret_path
from buidl.descriptor import P2WSHSortedMulti, parse_any_key_record
from buidl.hd import (
    calc_num_valid_seedpicker_checksums,
    calc_valid_seedpicker_checksums,
    HDPrivateKey,
    HDPublicKey,
    DEFAULT_P2WSH_PATH,
)
from buidl.libsec_status import is_libsec_enabled
from buidl.mnemonic import BIP39
from buidl.shamir import ShareSet
from buidl.psbt import MixedNetwork, PSBT


#####################################################################
# CLI UX
#####################################################################


# https://stackoverflow.com/questions/287871/how-to-print-colored-text-in-python
RESET_TERMINAL_COLOR = "\033[0m"


def blue_fg(string):
    return f"\033[34m{string}{RESET_TERMINAL_COLOR}"


def yellow_fg(string):
    return f"\033[93m{string}{RESET_TERMINAL_COLOR}"


def green_fg(string):
    return f"\033[32m{string}{RESET_TERMINAL_COLOR}"


def red_fg(string):
    return f"\033[31m{string}{RESET_TERMINAL_COLOR}"


def print_blue(string):
    print(blue_fg(string))


def print_yellow(string):
    print(yellow_fg(string))


def print_green(string):
    print(green_fg(string))


def print_red(string):
    print(red_fg(string))


def _get_buidl_version():
    try:
        return get_distribution("buidl").version
    except DistributionNotFound:
        return "Unknown"


def _get_int(prompt, default=20, minimum=0, maximum=None):
    while True:
        res = input(blue_fg(f"{prompt} [{default}]: ")).strip()
        if not res:
            res = str(default)
        try:
            res_int = int(res)
        except ValueError:
            print_red(f"{res} is not an integer")
            continue

        if maximum is not None:
            if not minimum <= res_int <= maximum:
                print_red(
                    f"Please pick a number between {minimum} and {maximum} (inclusive)"
                )
                continue

        if res_int < minimum:
            print_red(f"Please pick an integer >= {minimum}")
            continue

        return res_int


def _get_bool(prompt, default=True):
    if default is True:
        yn = "[Y/n]"
    else:
        yn = "[y/N]"

    while True:
        response_str = input(blue_fg(f"{prompt} {yn}: ")).strip().lower()
        if response_str == "":
            return default
        if response_str in ("n", "no"):
            return False
        if response_str in ("y", "yes"):
            return True
        print_red("Please choose either y or n")


def _get_string(prompt, default=""):
    return input(blue_fg(f"{prompt} [{default}]: ")).strip().lower() or default


def _get_path_string():
    while True:
        res = input(blue_fg('Path to use (should start with "m/"): ')).strip().lower()
        if not res.startswith("m/"):
            print_red(f'Invalid path "{res}" must start with "m/")')
            continue
        if res == "m/":
            # TODO: support this?
            print_red("Empty path (must have a depth of > 0): m/somenumberhere")
            continue
        for sub_path in res.split("/")[1:]:
            if sub_path.endswith("'") or sub_path.endswith("h"):
                # Trim trailing hardening indicator
                sub_path_cleaned = sub_path[:-1]
            else:
                sub_path_cleaned = sub_path
            try:
                int(sub_path_cleaned)
            except Exception:
                print_red(f"Invalid Path Section: {sub_path}")
                continue
        return res.replace("h", "'")


def _get_path(network):
    network_string = network.capitalize()
    default_path = DEFAULT_P2WSH_PATH[network]

    if _get_bool(
        prompt=f"Use default path ({default_path} for {network_string})",
        default=True,
    ):
        return default_path
    else:
        return _get_path_string()


def _get_confirmed_pw():
    while True:
        first = getpass(prompt=blue_fg("Enter custom passphrase: "))
        if first.strip() != first:
            print_red(
                "Leading/trailing spaces in passphrases are not supported. "
                "Please use an unambiguous passphrase."
            )
            continue
        second = getpass(prompt=blue_fg("Confirm custom passphrase: "))
        if first != second:
            print_red("Passphrases don't match, please try again.")
            continue
        return first


def _get_password():
    if _get_bool(
        prompt="Use a passphrase (advanced users only)?",
        default=False,
    ):
        return _get_confirmed_pw()
    else:
        return ""


class WordCompleter:
    def __init__(self, wordlist):
        self.wordlist = wordlist

    def complete(self, text, state):
        results = [x for x in self.wordlist if x.startswith(text)] + [None]
        return results[state] + " "


#####################################################################
# Output Descriptors
#####################################################################


def _get_p2wsh_sortedmulti():
    while True:
        output_record = input(
            blue_fg("Paste in your p2wsh output descriptors (account map): ")
        ).strip()
        try:
            return P2WSHSortedMulti.parse(output_record=output_record)
        except Exception as e:
            print_red(f"Could not parse output descriptor: {e}")
            if "wsh(sortedmulti(" not in output_record.lower():
                print_red("It should start with wsh(sortedmulti(... ")


#####################################################################
# Seedpicker
#####################################################################


def _get_bip39_firstwords():
    readline.parse_and_bind("tab: complete")
    old_completer = readline.get_completer()
    completer = WordCompleter(wordlist=BIP39)
    readline.set_completer(completer.complete)
    while True:
        fw = input(
            blue_fg("Enter the first 23 words of your BIP39 seed phrase: ")
        ).strip()
        fw_num = len(fw.split())

        if fw_num not in (11, 14, 17, 20, 23):
            # TODO: 11, 14, 17, or 20 word seed phrases also work but this is not documented as it's for advanced users
            print_red(
                f"You entered {fw_num} words. "
                "We recommend 23 words, but advanced users may enter 11, 14, 17 or 20 words."
            )
            continue

        all_words_valid = True
        for cnt, word in enumerate(fw.split()):
            if word not in BIP39:
                print_red(f"Word #{cnt+1} `{word}` is not a valid BIP39 word")
                all_words_valid = False
        if all_words_valid is False:
            continue

        readline.set_completer(old_completer)
        return fw


#####################################################################
# PSBT Signer
#####################################################################


def _get_psbt_obj():
    psbt_prompt = blue_fg(
        "Paste partially signed bitcoin transaction (PSBT) in base64 form: "
    )
    while True:
        psbt_b64 = input(psbt_prompt).strip()

        if not psbt_b64:
            continue

        try:
            # Attempt to infer network from BIP32 paths
            psbt_obj = PSBT.parse_base64(psbt_b64, network=None)
            if _get_bool(
                prompt=f"Transaction appears to be a {psbt_obj.network} transaction. Display as {psbt_obj.network}?",
                default=True,
            ):
                network = psbt_obj.network
            elif psbt_obj.network == "mainnet":
                network = "testnet"
            else:
                network = "mainnet"
            if psbt_obj.network != network:
                psbt_obj = PSBT.parse_base64(psbt_b64, network=network)

        except MixedNetwork:
            if _get_bool(
                prompt="Cannot infer PSBT network from BIP32 paths. Use Mainnet?",
                default=True,
            ):
                network = "mainnet"
            else:
                network = "testnet"
            psbt_obj = PSBT.parse_base64(psbt_b64, network=network)

        except Exception as e:
            print_red(f"Could not parse PSBT: {e}")
            continue

        # redundant but explicit
        if psbt_obj.validate() is not True:
            print_red("PSBT does not validate")
            continue

        return psbt_obj


def _get_bip39_seed(network):
    readline.parse_and_bind("tab: complete")
    old_completer = readline.get_completer()
    completer = WordCompleter(wordlist=BIP39)
    readline.set_completer(completer.complete)

    bip39_prompt = blue_fg("Enter your full BIP39 seed phrase: ")
    while True:
        seed_phrase = input(bip39_prompt).strip()
        seed_phrase_num = len(seed_phrase.split())

        if seed_phrase_num not in (12, 15, 18, 21, 24):
            # Other length seed phrases also work but this is not documented as it's for advanced users
            print_red(
                f"You entered {seed_phrase_num} words. "
                "By default seed phrases should be 24 words long (advanced users may enter seed phrases that are 12, 15, 18 or 21 words long)."
            )
            continue

        for cnt, word in enumerate(seed_phrase.split()):
            if word not in BIP39:
                print_red(f"Word #{cnt+1} ({word}) is not a valid BIP39 word")
                continue
        try:
            HDPrivateKey.from_mnemonic(mnemonic=seed_phrase, network=network)

            password = _get_password()
            hd_priv = HDPrivateKey.from_mnemonic(
                mnemonic=seed_phrase, network=network, password=password.encode()
            )
        except Exception as e:
            print_red(f"Invalid mnemonic: {e}")
            continue

        readline.set_completer(old_completer)
        return hd_priv


def _get_key_record(prompt):
    key_record_prompt = blue_fg(prompt)
    while True:
        key_record_str = input(key_record_prompt).strip()
        try:
            return parse_any_key_record(key_record_str=key_record_str)
        except ValueError as e:
            print_red(f"Could not parse entry: {e}")
            continue


def _get_units():
    # TODO: re-incorporate this into TX summary
    prompt = blue_fg("Units to diplay [BTC/sats]: ")
    while True:
        units = input(prompt).strip().lower()
        if units in ("", "btc", "btcs", "bitcoin", "bitcoins"):
            return "btc"
        if units in ("sat", "satoshis", "sats"):
            return "sats"


def _print_footgun_warning(custom_str=""):
    to_print = [
        "Running in SAFE mode.",
        "If you want to live dangerously, enter `(₿) advanced_mode` in the main menu.",
    ]
    if custom_str:
        to_print.append(custom_str)
    print_blue("\n".join(to_print) + "\n")


#####################################################################
# Command Line App Code Starts Here
#####################################################################


class MultiWallet(Cmd):
    ADVANCED_MODE = environ.get("ADVANCED_MODE", "").lower() in (
        "1",
        "t",
        "true",
        "y",
        "yes",
        "on",
    )
    intro = (
        "Welcome to multiwallet, the stateless multisig bitcoin wallet.\n"
        "This tool is free and there is NO WARRANTY OF ANY KIND.\n"
        f"You are currently in {'ADVANCED' if ADVANCED_MODE else 'SAFE'} mode.\n"
        "Type help or ? to list commands.\n"
    )
    prompt = "(₿) "  # the bitcoin symbol :)

    def __init__(self, stdin=sys.stdin, stdout=sys.stdout):
        Cmd.__init__(self, stdin=stdin, stdout=stdout)

    def do_generate_seed(self, arg):
        """Calculate bitcoin public and private key information from BIP39 words you draw out of a hat (using the seedpicker implementation)"""

        if not self.ADVANCED_MODE:
            _print_footgun_warning(
                "This will enable passphrases, custom paths, slip132 version bytes, and different checksum indices."
            )

        first_words = _get_bip39_firstwords()
        valid_checksums_generator = calc_valid_seedpicker_checksums(
            first_words=first_words
        )

        if self.ADVANCED_MODE:
            use_default_checksum = _get_bool(
                prompt="Use the default checksum index?", default=True
            )
        else:
            use_default_checksum = True

        if use_default_checksum:
            # This will be VERY fast
            last_word = next(valid_checksums_generator)
        else:
            num_valid_seedpicker_checksums = calc_num_valid_seedpicker_checksums(
                num_first_words=len(first_words.split())
            )

            to_print = f"Generating the {num_valid_seedpicker_checksums} valid checksum words for this seed phrase"
            if not is_libsec_enabled():
                to_print += " (this is ~10x faster if you install libsec)"
            to_print += ":\n"
            print_yellow(to_print)

            # This is slow, but will display progress in a UX-friendly way
            valid_checksum_words = []
            line = ""
            for i, word in enumerate(valid_checksums_generator):
                valid_checksum_words.append(word)
                current = f"{i}. {word}"
                if len(current) < 8:
                    current += "\t"
                line += f"{current}\t"
                if i % 4 == 3:
                    print_yellow("\t" + line)
                    line = ""
            print()
            index = _get_int(
                prompt="Choose one of the possible last words from above",
                default=0,
                minimum=0,
                maximum=len(valid_checksum_words) - 1,
            )
            last_word = valid_checksum_words[index]

        if self.ADVANCED_MODE:
            password = _get_password()
        else:
            password = ""

        if _get_bool(prompt="Use Mainnet?", default=False):
            network = "mainnet"
        else:
            network = "testnet"

        if self.ADVANCED_MODE:
            path_to_use = _get_path(network=network)
            use_slip132_version_byte = _get_bool(
                prompt="Encode with SLIP132 version byte?", default=True
            )
        else:
            path_to_use = None  # buidl will use default path
            use_slip132_version_byte = True

        hd_priv = HDPrivateKey.from_mnemonic(
            mnemonic=f"{first_words} {last_word}",
            password=password.encode(),
            network=network,
        )
        key_record = hd_priv.generate_p2wsh_key_record(
            bip32_path=path_to_use, use_slip132_version_byte=use_slip132_version_byte
        )
        print(yellow_fg("SECRET INFO") + red_fg(" (guard this VERY carefully)"))
        print_green(f"Last word: {last_word}")
        print_green(
            f"Full ({len(first_words.split()) + 1} word) mnemonic (including last word): {first_words + ' ' + last_word}"
        )
        if password:
            print_green(f"Passphrase: {password}")

        print_yellow(f"\nPUBLIC KEY INFO ({network})")
        print_yellow("Copy-paste this into Specter-Desktop:")
        print_green(key_record)

    def do_create_output_descriptors(self, arg):
        """Combine m-of-n public key records into a multisig output descriptor (account map)"""

        m_int = _get_int(
            prompt="How many signatures will be required to spend from this wallet?",
            default=2,
            minimum=1,
            maximum=15,
        )
        n_int = _get_int(
            prompt="How many total keys will be able to sign transaction from this wallet?",
            default=min(m_int + 1, 15),
            minimum=m_int,
            maximum=15,
        )
        key_records = []
        for cnt in range(n_int):
            prompt = f"Enter xpub key record #{cnt+1} (of {n_int}) in the format [deadbeef/path]xpub: "
            key_record_dict = _get_key_record(prompt)
            if key_record_dict.get("account_index") is None:
                # we have a partial dict
                key_record_dict["account_index"] = 0
                key_record_dict["xpub_parent"] = key_record_dict.pop("xpub")
            # Safety check
            if key_record_dict in key_records:
                # TODO: make this more performant
                print_red(
                    "ABORTING! Cannot use the same xpub key record twice in the same output descriptor."
                )
                return
            key_records.append(key_record_dict)

        sort_key_records = True
        if self.ADVANCED_MODE:
            sort_key_records = _get_bool(
                "Sort parent xpubs? This does not affect child addresses for sortedmulti, but alters bitcoin core descriptor checksum.",
                default=True,
            )

        p2wsh_sortedmulti_obj = P2WSHSortedMulti(
            quorum_m=m_int,
            key_records=key_records,
            sort_key_records=sort_key_records,
        )
        print_yellow("Your output descriptors are:\n")
        print_green(str(p2wsh_sortedmulti_obj))

    def do_blind_xpub(self, arg):
        """Blind an XPUB using a random BIP32 path. Experts only!"""

        # Prevent footgun-ing
        if not self.ADVANCED_MODE:
            warning_msg = (
                "Blinding an xpub must be performed correctly or you could lose funds!\n"
                "Multiwallet supports the custom BIP32 paths needed, but few HWWs can sign on these paths.\n"
                "Some can't even co-sign a multisig transaction with nonstandard BIP32 paths (even if their path is standard)"
            )
            _print_footgun_warning(warning_msg)
            return

        key_record_dict = _get_key_record(
            "Enter an xpub key record to blind in the format [deadbeef/path]xpub (any path will do): "
        )
        xfp_hex = key_record_dict["xfp"]
        starting_xpub = key_record_dict["xpub"]
        starting_path = key_record_dict["path"]

        depth = 4
        if not _get_bool("Use standard entropy parameters for blinding?", default=True):
            depth = _get_int(
                "EXPERTS ONLY: How deep of a BIP32 path do you want to blind this xpub with?",
                default=4,
                minimum=1,
                maximum=16,
            )
        # Get secret BIP32 path:
        secret_bip32_path = secure_secret_path(depth=depth)

        print_blue(f"Generating BIP32 path with {31*depth} bits of good entropy...")
        blinded_xpub_dict = blind_xpub(
            starting_xpub=starting_xpub,
            starting_path=starting_path,
            secret_path=secret_bip32_path,
        )
        blinded_full_path = blinded_xpub_dict["blinded_full_path"]
        blinded_child_xpub = blinded_xpub_dict["blinded_child_xpub"]
        blinded_key_record = f"[{xfp_hex}/{blinded_full_path[2:]}]{blinded_child_xpub}"

        explanation_msg = (
            "Here is a blinded xpub key record to upload to your Coordinator.\n"
            "Create a multisig wallet with this blinded xpub and it will become a part of your account map (output descriptors).\n"
        )
        print_yellow(explanation_msg)

        print_green(blinded_key_record)

        warning_msg = (
            "\nImportant notes:",
            "  - Do NOT share this record with the holder of the seed phrase, or they will be able to unblind their key (potentially leaking privacy info about what it controls).",
            "  - Possession of this blinded xpub key record has privacy implications, but it CANNOT alone be used to sign bitcoin transactions.",
            "  - Possession of the original seed phrase (used to create the original xpub key record), CANNOT alone be used to sign bitcoin transactions.",
            "  - In order to spend from this blinded xpub, you must have BOTH the seed phrase AND the blinded xpub key record (which will be included in your account map before you can receive funds).\n",
        )
        print_yellow("\n".join(warning_msg))

    def do_shamir_recover_seed(self, arg):
        """Recover a seed from Shamir shares using a SLIP39-like standard"""
        # Prevent footgun-ing
        if not self.ADVANCED_MODE:
            warning_msg = (
                "This shamir share implementation is non-standard!\n"
                "If you're trying to recover from SLIP39 seed generated elsewhere, it probably won't work."
            )
            _print_footgun_warning(warning_msg)
            return

        share_mnemonics = []
        while True:
            share_phrase = input(
                blue_fg("Enter a SLIP39 Shamir share (blank to end): ")
            ).strip()
            if share_phrase == "":
                break
            share_mnemonics.append(share_phrase)
        passphrase = b""
        if self.ADVANCED_MODE:
            has_passphrase = _get_bool("Is there a passphrase?", default=False)
            if has_passphrase:
                passphrase = _get_confirmed_pw().encode("ascii")
        try:
            mnemonic = ShareSet.recover_mnemonic(share_mnemonics, passphrase=passphrase)
            print_green(f"Here is your recovered mnemonic:\n\n{mnemonic}")
        except (TypeError, ValueError, SyntaxError) as e:
            print_red(e)

    def do_shamir_split_seed(self, arg):
        """Split a seed to Shamir shares using a SLIP39-like standard"""
        # Prevent footgun-ing
        if not self.ADVANCED_MODE:
            warning_msg = (
                "This shamir share implementation is non-standard!\n"
                "It has the benefit of being reversible, meaning it accepts standard BIP39 seeds and can recreate them later from Shamir Shares, the SSSS protocol is not compatible with SLIP39."
                "However, once you recover your original BIP39 seed phrase (not possible with SLIP39), you can put that into any hardware wallet."
            )
            _print_footgun_warning(warning_msg)
            return

        mnemonic = input(blue_fg("Enter a BIP39 seed phrase: ")).strip()
        k = _get_int(
            "How many Shamir shares should be required to recover the seed phrase?",
            default=2,
            minimum=1,
            maximum=16,
        )
        default_n = 2 * k - 1 if 2 * k - 1 < 16 else 16
        n = _get_int(
            "How many Shamir shares do you want to generate?",
            default=default_n,
            minimum=k,
            maximum=16,
        )
        passphrase = b""
        # Max possible combos is 12,870 (with 8-of-16)
        testing_limit = 12870
        if self.ADVANCED_MODE:
            add_passphrase = _get_bool(
                "Do you want to add a passphrase?", default=False
            )
            if add_passphrase:
                passphrase = _get_confirmed_pw().encode("ascii")
            if _get_bool(
                "Do you want to limit how many of these shares you test (for faster results)?",
                default=False,
            ):
                testing_limit = _get_int(
                    "How many combinations would you like to test?",
                    default=1000,
                    minimum=k,
                    maximum=testing_limit,
                )
        shares = ShareSet.generate_shares(mnemonic, k, n, passphrase=passphrase)
        # test the shares (cap at 1000 combinations)
        print_yellow(
            "Testing share combinations to be certain they will recover your seed phrase"
        )
        for cnt, combo in enumerate(combinations(shares, k)):
            if cnt > testing_limit:
                break
            calculated = ShareSet.recover_mnemonic(combo, passphrase)
            if calculated != mnemonic:
                # we should never reach this line
                raise RuntimeError(f"Bad shares {calculated} for mnemonic {mnemonic}")
            if cnt % 10 == 0:
                print(".", end="", flush=True)

        print_yellow(
            f"\nSuccesfully tested {'ALL ' if testing_limit == 12870 else ''}{cnt} combinations"
        )

        print("\n")
        share_mnemonics = "\n\n".join(shares)
        if passphrase:
            additional = " AND your passphrase"
        else:
            additional = ""
        prompt = f"You will need {k} of these {n} share phrases{additional} to recover your seed phrase:\n\n{share_mnemonics}"
        print_green(prompt)

    def do_validate_address(self, arg):
        """Verify receive addresses for a multisig wallet using output descriptors (from Specter-Desktop)"""
        p2wsh_sortedmulti_obj = _get_p2wsh_sortedmulti()
        limit = _get_int(
            prompt="Limit of addresses to display",
            # This is slow without libsecp256k1:
            default=25 if is_libsec_enabled() else 5,
            minimum=1,
        )
        offset = _get_int(
            prompt="Offset of addresses to display",
            default=0,
            minimum=0,
        )

        # Only advanced users should consider seeing change addresses
        is_change = False
        if self.ADVANCED_MODE:
            is_change = not _get_bool(
                prompt="Display receive addresses? `N` to display change addresses instead.",
                default=True,
            )

        to_print = f"{p2wsh_sortedmulti_obj.m_of_n} Multisig {'Change' if is_change else 'Receive'} Addresses"
        if not is_libsec_enabled():
            to_print += "\n(this is ~100x faster if you install libsec)"
        print_yellow(to_print + ":")
        for cnt in range(limit):
            offset_to_use = offset + cnt
            address = p2wsh_sortedmulti_obj.get_address(
                is_change=is_change,
                offset=offset_to_use,
            )
            print_green(f"#{offset_to_use}: {address}")

    def do_sign_transaction(self, arg):
        """
        (Co)sign a multisig PSBT using 1 of your BIP39 seed phrases.
        Can also be used to just inspect a PSBT and not sign it.

        Note: This tool ONLY supports transactions with the following constraints:
          - We sign ALL inputs and they belong to the same multisig wallet (quorum + pubkeys).
          - There can only be 1 output (sweep transaction) or 2 outputs (spend + change).
          - If there is change, we validate it belongs to the same multisig wallet as all inputs.
        """

        psbt_obj = _get_psbt_obj()

        if psbt_obj.hd_pubs:
            # if the PSBT included hd_pubs, then buidl will build the hdpubkey_map automatically from that
            hdpubkey_map = {}
        else:
            # ask for output descriptors to use to build hdpubkey_map
            print_blue(
                "PSBT doesn't include enough info to guess your account map (for validation)."
            )
            p2wsh_sortedmulti_obj = _get_p2wsh_sortedmulti()

            hdpubkey_map = {}
            for key_record in p2wsh_sortedmulti_obj.key_records:
                hdpubkey_map[key_record["xfp"]] = HDPublicKey.parse(
                    key_record["xpub_parent"]
                )

        psbt_described = psbt_obj.describe_basic_multisig(
            hdpubkey_map=hdpubkey_map,
        )

        # Gather TX info and validate
        print_yellow(psbt_described["tx_summary_text"])

        if _get_bool(prompt="In Depth Transaction View?", default=False):
            to_print = []
            to_print.append("DETAILED VIEW")
            to_print.append(f"TXID: {psbt_obj.tx_obj.id()}")
            to_print.append("-" * 80)
            to_print.append(f"{len(psbt_described['inputs_desc'])} Input(s):")
            for cnt, input_desc in enumerate(psbt_described["inputs_desc"]):
                to_print.append(f"  Input #{cnt}")
                for k, v in input_desc.items():
                    if k == "sats":
                        # Comma separate ints
                        val = f"{v:,}"
                    else:
                        val = v
                    to_print.append(f"    {k}: {val}")
            to_print.append("-" * 80)
            to_print.append(f"{len(psbt_described['outputs_desc'])} Output(s):")
            for cnt, output_desc in enumerate(psbt_described["outputs_desc"]):
                to_print.append(f"  Output #{cnt}")
                for k, v in output_desc.items():
                    if k == "sats":
                        # Comma separate ints
                        val = f"{v:,}"
                    else:
                        val = v
                    to_print.append(f"    {k}: {val}")
            print_yellow("\n".join(to_print))

        if not _get_bool(prompt="Sign this transaction?", default=True):
            print_yellow(f"Transaction {psbt_obj.tx_obj.id()} NOT signed")
            return

        hd_priv = _get_bip39_seed(network=psbt_obj.network)
        root_paths_for_seed = psbt_described["root_paths"][hd_priv.fingerprint().hex()]
        private_keys = [
            hd_priv.traverse(root_path).private_key for root_path in root_paths_for_seed
        ]

        if psbt_obj.sign_with_private_keys(private_keys) is True:
            print()
            print_yellow("Signed PSBT to broadcast:\n")
            print_green(psbt_obj.serialize_base64())
        else:
            print_red("ERROR: Could NOT sign PSBT!")
            return

    def do_convert_descriptors_to_caravan(self, arg):
        """
        Convert bitcoin core output descriptors to caravan import
        """
        p2wsh_sortedmulti_obj = _get_p2wsh_sortedmulti()

        wallet_name = _get_string("Enter wallet name", default="p2wsh wallet")

        key_record_names = []
        if _get_bool("Give human name to each key record?", default=False):
            for cnt, kr in enumerate(p2wsh_sortedmulti_obj.key_records):
                kr_name = _get_string(
                    f"Enter key record name for {kr['xfp']}", default=f"Seed #{cnt+1}"
                )
                key_record_names.append(kr_name)

        caravan_json = p2wsh_sortedmulti_obj.caravan_export(
            wallet_name=wallet_name, key_record_names=key_record_names
        )
        print_yellow("Output descriptor as Caravan import (save this to a file):")
        print_green(caravan_json)

    def do_toggle_advanced_mode(self, arg):
        """
        Toggle advanced mode features like passphrases, different BIP39 seed checksums, non-standard BIP32 paths, xpub blinding, shamir's secret sharing scheme, etc.
        WARNING: these features are for advanced users and could lead to loss of funds.
        """
        if self.ADVANCED_MODE:
            self.ADVANCED_MODE = False
            print_yellow("SAFE mode set, your training wheels have been restored!")
        else:
            self.ADVANCED_MODE = True
            print_yellow("ADVANCED mode set, don't mess up!")

    def do_version_info(self, arg):
        """Print program settings for debug purposes"""

        to_print = [
            f"buidl Version: {_get_buidl_version()}",
            f"Multiwallet Mode: {'Advanced' if self.ADVANCED_MODE else 'Basic'}",
            f"Python Version: {sys.version_info}",
            f"Platform: {platform()}",
            f"libsecp256k1 Configured: {is_libsec_enabled()}",
        ]
        print_yellow("\n".join(to_print))

    def do_exit(self, arg):
        """Exit Program"""
        print_yellow("\nNo data saved")
        return True


if __name__ == "__main__":
    try:
        MultiWallet().cmdloop()
    except KeyboardInterrupt:
        print_yellow("\nNo data saved")



import math

from buidl.block import Block

from buidl.helper import (
    bytes_to_bit_field,
    little_endian_to_int,
    merkle_parent,
    read_varint,
)


class MerkleTree:
    def __init__(self, total):
        self.total = total
        # compute max depth math.ceil(math.log(self.total, 2))
        self.max_depth = math.ceil(math.log(self.total, 2))
        # initialize the nodes property to hold the actual tree
        self.nodes = []
        # loop over the number of levels (max_depth+1)
        for depth in range(self.max_depth + 1):
            # the number of items at this depth is
            # math.ceil(self.total / 2**(self.max_depth - depth))
            num_items = math.ceil(self.total / 2 ** (self.max_depth - depth))
            # create this level's hashes list with the right number of items
            level_hashes = [None] * num_items
            # append this level's hashes to the merkle tree
            self.nodes.append(level_hashes)
        # set the pointer to the root (depth=0, index=0)
        self.current_depth = 0
        self.current_index = 0
        self.proved_txs = []

    def __repr__(self):
        result = []
        for depth, level in enumerate(self.nodes):
            items = []
            for index, h in enumerate(level):
                if h is None:
                    short = "None"
                else:
                    short = "{}...".format(h.hex()[:8])
                if depth == self.current_depth and index == self.current_index:
                    items.append("*{}*".format(short[:-2]))
                else:
                    items.append("{}".format(short))
            result.append(", ".join(items))
        return "\n".join(result)

    def up(self):
        # reduce depth by 1 and halve the index
        self.current_depth -= 1
        self.current_index //= 2

    def left(self):
        # increase depth by 1 and double the index
        self.current_depth += 1
        self.current_index *= 2

    def right(self):
        # increase depth by 1 and double the index + 1
        self.current_depth += 1
        self.current_index = self.current_index * 2 + 1

    def root(self):
        return self.nodes[0][0]

    def set_current_node(self, value):
        self.nodes[self.current_depth][self.current_index] = value

    def get_current_node(self):
        return self.nodes[self.current_depth][self.current_index]

    def get_left_node(self):
        return self.nodes[self.current_depth + 1][self.current_index * 2]

    def get_right_node(self):
        return self.nodes[self.current_depth + 1][self.current_index * 2 + 1]

    def is_leaf(self):
        return self.current_depth == self.max_depth

    def right_exists(self):
        return len(self.nodes[self.current_depth + 1]) > self.current_index * 2 + 1

    def populate_tree(self, flag_bits, hashes):
        # populate until we have the root
        while self.root() is None:
            # if we are a leaf, we know this position's hash
            if self.is_leaf():
                # get the next bit from flag_bits: flag_bits.pop(0)
                flag_bit = flag_bits.pop(0)
                # get the current hash from hashes: hashes.pop(0)
                current_hash = hashes.pop(0)
                # set the current node in the merkle tree to the current hash
                self.set_current_node(current_hash)
                # if our flag bit is 1, add to the self.proved_txs array
                if flag_bit == 1:
                    self.proved_txs.append(current_hash[::-1])
                # go up a level
                self.up()
            # else
            else:
                # get the left hash
                left_hash = self.get_left_node()
                # if we don't have the left hash
                if left_hash is None:
                    # if the next flag bit is 0, the next hash is our current node
                    if flag_bits.pop(0) == 0:
                        # set the current node to be the next hash
                        self.set_current_node(hashes.pop(0))
                        # sub-tree doesn't need calculation, go up
                        self.up()
                    # else
                    else:
                        # go to the left node
                        self.left()
                elif self.right_exists():
                    # get the right hash
                    right_hash = self.get_right_node()
                    # if we don't have the right hash
                    if right_hash is None:
                        # go to the right node
                        self.right()
                    # else
                    else:
                        # combine the left and right hashes
                        self.set_current_node(merkle_parent(left_hash, right_hash))
                        # we've completed this sub-tree, go up
                        self.up()
                # else
                else:
                    # combine the left hash twice
                    self.set_current_node(merkle_parent(left_hash, left_hash))
                    # we've completed this sub-tree, go up
                    self.up()
        if len(hashes) != 0:
            raise RuntimeError("hashes not all consumed {}".format(len(hashes)))
        for flag_bit in flag_bits:
            if flag_bit != 0:
                raise RuntimeError("flag bits not all consumed")


class MerkleBlock:
    command = b"merkleblock"

    def __init__(self, header, total, hashes, flags):
        self.header = header
        self.total = total
        self.hashes = hashes
        self.flags = flags
        self.merkle_tree = None

    def __repr__(self):
        result = "{}\n".format(self.total)
        for h in self.hashes:
            result += "\t{}\n".format(h.hex())
        result += "{}".format(self.flags.hex())

    def hash(self):
        return self.header.hash()

    def id(self):
        return self.header.id()

    @classmethod
    def parse(cls, s):
        """Takes a byte stream and parses a merkle block. Returns a Merkle Block object"""
        # s.read(n) will read n bytes from the stream
        # header - use Block.parse_header with the stream
        header = Block.parse_header(s)
        # total number of transactions (4 bytes, little endian)
        total = little_endian_to_int(s.read(4))
        # number of hashes is a varint
        num_txs = read_varint(s)
        # initialize the hashes array
        hashes = []
        # loop through the number of hashes times
        for _ in range(num_txs):
            # each hash is 32 bytes, little endian
            hashes.append(s.read(32)[::-1])
        # get the length of the flags field as a varint
        flags_length = read_varint(s)
        # read the flags field
        flags = s.read(flags_length)
        # initialize class
        return cls(header, total, hashes, flags)

    def is_valid(self):
        """Verifies whether the merkle tree information validates to the merkle root"""
        # use bytes_to_bit_field on self.flags to get the flag_bits
        flag_bits = bytes_to_bit_field(self.flags)
        # set hashes to be the reversed hashes of everything in self.hashes
        hashes = [h[::-1] for h in self.hashes]
        # initialize the merkle tree with self.total
        self.merkle_tree = MerkleTree(self.total)
        # populate_tree with flag_bits and hashes
        self.merkle_tree.populate_tree(flag_bits, hashes)
        # check if the computed root [::-1] is the same as the merkle root
        return self.merkle_tree.root()[::-1] == self.header.merkle_root

    def proved_txs(self):
        """Returns the list of proven transactions from the Merkle block"""
        if self.merkle_tree is None:
            return []
        else:
            return self.merkle_tree.proved_txs



import hashlib

from buidl.ecc import (
    S256Point,
    Signature,
)

from buidl.helper import (
    hash160,
    hash256,
)


def encode_num(num):
    if num == 0:
        return b""
    abs_num = abs(num)
    negative = num < 0
    result = bytearray()
    while abs_num:
        result.append(abs_num & 0xFF)
        abs_num >>= 8
    # if the top bit is set,
    # for negative numbers we ensure that the top bit is set
    # for positive numbers we ensure that the top bit is not set
    if result[-1] & 0x80:
        if negative:
            result.append(0x80)
        else:
            result.append(0)
    elif negative:
        result[-1] |= 0x80
    return bytes(result)


def decode_num(element):
    if element == b"":
        return 0
    # reverse for big endian
    big_endian = element[::-1]
    # top bit being 1 means it's negative
    if big_endian[0] & 0x80:
        negative = True
        result = big_endian[0] & 0x7F
    else:
        negative = False
        result = big_endian[0]
    for c in big_endian[1:]:
        result <<= 8
        result += c
    if negative:
        return -result
    else:
        return result


def op_0(stack):
    stack.append(encode_num(0))
    return True


def op_1negate(stack):
    stack.append(encode_num(-1))
    return True


def op_1(stack):
    stack.append(encode_num(1))
    return True


def op_2(stack):
    stack.append(encode_num(2))
    return True


def op_3(stack):
    stack.append(encode_num(3))
    return True


def op_4(stack):
    stack.append(encode_num(4))
    return True


def op_5(stack):
    stack.append(encode_num(5))
    return True


def op_6(stack):
    stack.append(encode_num(6))
    return True


def op_7(stack):
    stack.append(encode_num(7))
    return True


def op_8(stack):
    stack.append(encode_num(8))
    return True


def op_9(stack):
    stack.append(encode_num(9))
    return True


def op_10(stack):
    stack.append(encode_num(10))
    return True


def op_11(stack):
    stack.append(encode_num(11))
    return True


def op_12(stack):
    stack.append(encode_num(12))
    return True


def op_13(stack):
    stack.append(encode_num(13))
    return True


def op_14(stack):
    stack.append(encode_num(14))
    return True


def op_15(stack):
    stack.append(encode_num(15))
    return True


def op_16(stack):
    stack.append(encode_num(16))
    return True


def op_nop(stack):
    return True


def op_if(stack, items):
    if len(stack) < 1:
        return False
    # go through and re-make the items array based on the top stack element
    true_items = []
    false_items = []
    current_array = true_items
    found = False
    num_endifs_needed = 1
    while len(items) > 0:
        item = items.pop(0)
        if item in (99, 100):
            # nested if, we have to go another endif
            num_endifs_needed += 1
            current_array.append(item)
        elif num_endifs_needed == 1 and item == 103:
            current_array = false_items
        elif item == 104:
            if num_endifs_needed == 1:
                found = True
                break
            else:
                num_endifs_needed -= 1
                current_array.append(item)
        else:
            current_array.append(item)
    if not found:
        return False
    element = stack.pop()
    if decode_num(element) == 0:
        items[:0] = false_items
    else:
        items[:0] = true_items
    return True


def op_notif(stack, items):
    if len(stack) < 1:
        return False
    # go through and re-make the items array based on the top stack element
    true_items = []
    false_items = []
    current_array = true_items
    found = False
    num_endifs_needed = 1
    while len(items) > 0:
        item = items.pop(0)
        if item in (99, 100):
            # nested if, we have to go another endif
            num_endifs_needed += 1
            current_array.append(item)
        elif num_endifs_needed == 1 and item == 103:
            current_array = false_items
        elif item == 104:
            if num_endifs_needed == 1:
                found = True
                break
            else:
                num_endifs_needed -= 1
                current_array.append(item)
        else:
            current_array.append(item)
    if not found:
        return False
    element = stack.pop()
    if decode_num(element) == 0:
        items[:0] = true_items
    else:
        items[:0] = false_items
    return True


def op_verify(stack):
    if len(stack) < 1:
        return False
    element = stack.pop()
    if decode_num(element) == 0:
        return False
    return True


def op_return(stack):
    return False


def op_toaltstack(stack, altstack):
    if len(stack) < 1:
        return False
    altstack.append(stack.pop())
    return True


def op_fromaltstack(stack, altstack):
    if len(altstack) < 1:
        return False
    stack.append(altstack.pop())
    return True


def op_2drop(stack):
    if len(stack) < 2:
        return False
    stack.pop()
    stack.pop()
    return True


def op_2dup(stack):
    if len(stack) < 2:
        return False
    stack.extend(stack[-2:])
    return True


def op_3dup(stack):
    if len(stack) < 3:
        return False
    stack.extend(stack[-3:])
    return True


def op_2over(stack):
    if len(stack) < 4:
        return False
    stack.extend(stack[-4:-2])
    return True


def op_2rot(stack):
    if len(stack) < 6:
        return False
    stack.extend(stack[-6:-4])
    return True


def op_2swap(stack):
    if len(stack) < 4:
        return False
    stack[-4:] = stack[-2:] + stack[-4:-2]
    return True


def op_ifdup(stack):
    if len(stack) < 1:
        return False
    if decode_num(stack[-1]) != 0:
        stack.append(stack[-1])
    return True


def op_depth(stack):
    stack.append(encode_num(len(stack)))
    return True


def op_drop(stack):
    if len(stack) < 1:
        return False
    stack.pop()
    return True


def op_dup(stack):
    if len(stack) < 1:
        return False
    stack.append(stack[-1])
    return True


def op_nip(stack):
    if len(stack) < 2:
        return False
    stack[-2:] = stack[-1:]
    return True


def op_over(stack):
    if len(stack) < 2:
        return False
    stack.append(stack[-2])
    return True


def op_pick(stack):
    if len(stack) < 1:
        return False
    n = decode_num(stack.pop())
    if len(stack) < n + 1:
        return False
    stack.append(stack[-n - 1])
    return True


def op_roll(stack):
    if len(stack) < 1:
        return False
    n = decode_num(stack.pop())
    if len(stack) < n + 1:
        return False
    if n == 0:
        return True
    stack.append(stack.pop(-n - 1))
    return True


def op_rot(stack):
    if len(stack) < 3:
        return False
    stack.append(stack.pop(-3))
    return True


def op_swap(stack):
    if len(stack) < 2:
        return False
    stack.append(stack.pop(-2))
    return True


def op_tuck(stack):
    if len(stack) < 2:
        return False
    stack.insert(-2, stack[-1])
    return True


def op_size(stack):
    if len(stack) < 1:
        return False
    stack.append(encode_num(len(stack[-1])))
    return True


def op_equal(stack):
    if len(stack) < 2:
        return False
    element1 = stack.pop()
    element2 = stack.pop()
    if element1 == element2:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_equalverify(stack):
    return op_equal(stack) and op_verify(stack)


def op_1add(stack):
    if len(stack) < 1:
        return False
    element = decode_num(stack.pop())
    stack.append(encode_num(element + 1))
    return True


def op_1sub(stack):
    if len(stack) < 1:
        return False
    element = decode_num(stack.pop())
    stack.append(encode_num(element - 1))
    return True


def op_negate(stack):
    if len(stack) < 1:
        return False
    element = decode_num(stack.pop())
    stack.append(encode_num(-element))
    return True


def op_abs(stack):
    if len(stack) < 1:
        return False
    element = decode_num(stack.pop())
    if element < 0:
        stack.append(encode_num(-element))
    else:
        stack.append(encode_num(element))
    return True


def op_not(stack):
    if len(stack) < 1:
        return False
    element = stack.pop()
    if decode_num(element) == 0:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_0notequal(stack):
    if len(stack) < 1:
        return False
    element = stack.pop()
    if decode_num(element) == 0:
        stack.append(encode_num(0))
    else:
        stack.append(encode_num(1))
    return True


def op_add(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    stack.append(encode_num(element1 + element2))
    return True


def op_sub(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    stack.append(encode_num(element2 - element1))
    return True


def op_booland(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element1 and element2:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_boolor(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element1 or element2:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_numequal(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element1 == element2:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_numequalverify(stack):
    return op_numequal(stack) and op_verify(stack)


def op_numnotequal(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element1 == element2:
        stack.append(encode_num(0))
    else:
        stack.append(encode_num(1))
    return True


def op_lessthan(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element2 < element1:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_greaterthan(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element2 > element1:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_lessthanorequal(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element2 <= element1:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_greaterthanorequal(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element2 >= element1:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_min(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element1 < element2:
        stack.append(encode_num(element1))
    else:
        stack.append(encode_num(element2))
    return True


def op_max(stack):
    if len(stack) < 2:
        return False
    element1 = decode_num(stack.pop())
    element2 = decode_num(stack.pop())
    if element1 > element2:
        stack.append(encode_num(element1))
    else:
        stack.append(encode_num(element2))
    return True


def op_within(stack):
    if len(stack) < 3:
        return False
    maximum = decode_num(stack.pop())
    minimum = decode_num(stack.pop())
    element = decode_num(stack.pop())
    if element >= minimum and element < maximum:
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_ripemd160(stack):
    if len(stack) < 1:
        return False
    element = stack.pop()
    stack.append(hashlib.new("ripemd160", element).digest())
    return True


def op_sha1(stack):
    if len(stack) < 1:
        return False
    element = stack.pop()
    stack.append(hashlib.sha1(element).digest())
    return True


def op_sha256(stack):
    if len(stack) < 1:
        return False
    element = stack.pop()
    stack.append(hashlib.sha256(element).digest())
    return True


def op_hash160(stack):
    # check to see if there's at least 1 element
    if len(stack) < 1:
        return False
    # get the element on the top with stack.pop()
    element = stack.pop()
    # add the hash160 of the element to the end of the stack
    h160 = hash160(element)
    stack.append(h160)
    return True


def op_hash256(stack):
    if len(stack) < 1:
        return False
    element = stack.pop()
    stack.append(hash256(element))
    return True


def op_checksig(stack, z):
    # check to see if there's at least 2 elements
    if len(stack) < 2:
        return False
    # get the sec_pubkey with stack.pop()
    sec_pubkey = stack.pop()
    # get the der_signature with stack.pop()[:-1] (last byte is removed)
    der_signature = stack.pop()[:-1]
    # parse the sec format pubkey with S256Point
    point = S256Point.parse(sec_pubkey)
    # parse the der format signature with Signature
    sig = Signature.parse(der_signature)
    # verify using the point, z and signature
    # if verified add encode_num(1) to the end, otherwise encode_num(0)
    if point.verify(z, sig):
        stack.append(encode_num(1))
    else:
        stack.append(encode_num(0))
    return True


def op_checksigverify(stack, z):
    return op_checksig(stack, z) and op_verify(stack)


def op_checkmultisig(stack, z):
    if len(stack) < 1:
        return False
    n = decode_num(stack.pop())
    if len(stack) < n + 1:
        return False
    sec_pubkeys = []
    for _ in range(n):
        sec_pubkeys.append(stack.pop())
    m = decode_num(stack.pop())
    if len(stack) < m + 1:
        return False
    der_signatures = []
    for _ in range(m):
        # signature is assumed to be using SIGHASH_ALL
        der_signatures.append(stack.pop()[:-1])
    # OP_CHECKMULTISIG bug
    stack.pop()
    try:
        # parse the sec pubkeys into an array of points
        points = [S256Point.parse(sec) for sec in sec_pubkeys]
        # parse the der_signatures into an array of signatures
        sigs = [Signature.parse(der) for der in der_signatures]
        # loop through the signatures
        for sig in sigs:
            # bail early if we don't have any points left
            if len(points) == 0:
                print("signatures no good or not in right order")
                return False
            # while we have points
            while points:
                # get the point at the front (points.pop(0))
                point = points.pop(0)
                # see if this point can verify this sig with this z
                if point.verify(z, sig):
                    # break if so, this sig is valid!
                    break
        # if we made it this far, we have to add a 1 to the stack
        # use encode_num(1)
        stack.append(encode_num(1))
    except (ValueError, SyntaxError):
        return False
    return True


def op_checkmultisigverify(stack, z):
    return op_checkmultisig(stack, z) and op_verify(stack)


def op_checklocktimeverify(stack, locktime, sequence):
    if sequence == 0xFFFFFFFF:
        return False
    if len(stack) < 1:
        return False
    element = decode_num(stack[-1])
    if element < 0:
        return False
    if element < 500000000 and locktime > 500000000:
        return False
    if locktime < element:
        return False
    return True


def op_checksequenceverify(stack, version, sequence):
    if sequence & (1 << 31) == (1 << 31):
        return False
    if len(stack) < 1:
        return False
    element = decode_num(stack[-1])
    if element < 0:
        return False
    if element & (1 << 31) == (1 << 31):
        if version < 2:
            return False
        elif sequence & (1 << 31) == (1 << 31):
            return False
        elif element & (1 << 22) != sequence & (1 << 22):
            return False
        elif element & 0xFFFF > sequence & 0xFFFF:
            return False
    return True


def number_to_op_code_byte(n):
    """Returns the OP code for a particular number"""
    if n < -1 or n > 16:
        raise ValueError("Not a valid OP code")
    if n > 0:
        return bytes([0x50 + n])
    elif n == 0:
        return b"\x00"
    elif n == -1:
        return b"\x4f"


def number_to_op_code(n):
    if n < -1 or n > 16:
        raise ValueError("Not a valid OP code")

    if n == 0:
        return 0

    return n + 80


def op_code_to_number(op_code):
    """Returns the n for a particular OP code"""
    if op_code not in (
        0,
        79,
        80,
        81,
        82,
        83,
        84,
        85,
        86,
        87,
        88,
        89,
        90,
        91,
        92,
        93,
        94,
        95,
        96,
    ):
        raise ValueError("Not a valid OP code")
    if op_code == 0:
        return 0
    else:
        return op_code - 80


OP_CODE_FUNCTIONS = {
    0: op_0,
    79: op_1negate,
    81: op_1,
    82: op_2,
    83: op_3,
    84: op_4,
    85: op_5,
    86: op_6,
    87: op_7,
    88: op_8,
    89: op_9,
    90: op_10,
    91: op_11,
    92: op_12,
    93: op_13,
    94: op_14,
    95: op_15,
    96: op_16,
    97: op_nop,
    99: op_if,
    100: op_notif,
    105: op_verify,
    106: op_return,
    107: op_toaltstack,
    108: op_fromaltstack,
    109: op_2drop,
    110: op_2dup,
    111: op_3dup,
    112: op_2over,
    113: op_2rot,
    114: op_2swap,
    115: op_ifdup,
    116: op_depth,
    117: op_drop,
    118: op_dup,
    119: op_nip,
    120: op_over,
    121: op_pick,
    122: op_roll,
    123: op_rot,
    124: op_swap,
    125: op_tuck,
    130: op_size,
    135: op_equal,
    136: op_equalverify,
    139: op_1add,
    140: op_1sub,
    143: op_negate,
    144: op_abs,
    145: op_not,
    146: op_0notequal,
    147: op_add,
    148: op_sub,
    154: op_booland,
    155: op_boolor,
    156: op_numequal,
    157: op_numequalverify,
    158: op_numnotequal,
    159: op_lessthan,
    160: op_greaterthan,
    161: op_lessthanorequal,
    162: op_greaterthanorequal,
    163: op_min,
    164: op_max,
    165: op_within,
    166: op_ripemd160,
    167: op_sha1,
    168: op_sha256,
    169: op_hash160,
    170: op_hash256,
    172: op_checksig,
    173: op_checksigverify,
    174: op_checkmultisig,
    175: op_checkmultisigverify,
    176: op_nop,
    177: op_checklocktimeverify,
    178: op_checksequenceverify,
    179: op_nop,
    180: op_nop,
    181: op_nop,
    182: op_nop,
    183: op_nop,
    184: op_nop,
    185: op_nop,
}

OP_CODE_NAMES = {
    0: "OP_0",
    76: "OP_PUSHDATA1",
    77: "OP_PUSHDATA2",
    78: "OP_PUSHDATA4",
    79: "OP_1NEGATE",
    81: "OP_1",
    82: "OP_2",
    83: "OP_3",
    84: "OP_4",
    85: "OP_5",
    86: "OP_6",
    87: "OP_7",
    88: "OP_8",
    89: "OP_9",
    90: "OP_10",
    91: "OP_11",
    92: "OP_12",
    93: "OP_13",
    94: "OP_14",
    95: "OP_15",
    96: "OP_16",
    97: "OP_NOP",
    99: "OP_IF",
    100: "OP_NOTIF",
    103: "OP_ELSE",
    104: "OP_ENDIF",
    105: "OP_VERIFY",
    106: "OP_RETURN",
    107: "OP_TOALTSTACK",
    108: "OP_FROMALTSTACK",
    109: "OP_2DROP",
    110: "OP_2DUP",
    111: "OP_3DUP",
    112: "OP_2OVER",
    113: "OP_2ROT",
    114: "OP_2SWAP",
    115: "OP_IFDUP",
    116: "OP_DEPTH",
    117: "OP_DROP",
    118: "OP_DUP",
    119: "OP_NIP",
    120: "OP_OVER",
    121: "OP_PICK",
    122: "OP_ROLL",
    123: "OP_ROT",
    124: "OP_SWAP",
    125: "OP_TUCK",
    130: "OP_SIZE",
    135: "OP_EQUAL",
    136: "OP_EQUALVERIFY",
    139: "OP_1ADD",
    140: "OP_1SUB",
    143: "OP_NEGATE",
    144: "OP_ABS",
    145: "OP_NOT",
    146: "OP_0NOTEQUAL",
    147: "OP_ADD",
    148: "OP_SUB",
    154: "OP_BOOLAND",
    155: "OP_BOOLOR",
    156: "OP_NUMEQUAL",
    157: "OP_NUMEQUALVERIFY",
    158: "OP_NUMNOTEQUAL",
    159: "OP_LESSTHAN",
    160: "OP_GREATERTHAN",
    161: "OP_LESSTHANOREQUAL",
    162: "OP_GREATERTHANOREQUAL",
    163: "OP_MIN",
    164: "OP_MAX",
    165: "OP_WITHIN",
    166: "OP_RIPEMD160",
    167: "OP_SHA1",
    168: "OP_SHA256",
    169: "OP_HASH160",
    170: "OP_HASH256",
    171: "OP_CODESEPARATOR",
    172: "OP_CHECKSIG",
    173: "OP_CHECKSIGVERIFY",
    174: "OP_CHECKMULTISIG",
    175: "OP_CHECKMULTISIGVERIFY",
    176: "OP_NOP1",
    177: "OP_CHECKLOCKTIMEVERIFY",
    178: "OP_CHECKSEQUENCEVERIFY",
    179: "OP_NOP4",
    180: "OP_NOP5",
    181: "OP_NOP6",
    182: "OP_NOP7",
    183: "OP_NOP8",
    184: "OP_NOP9",
    185: "OP_NOP10",
}



from collections import defaultdict

from buidl.hd import get_unhardened_child_path, HDPublicKey
from buidl.psbt import MixedNetwork, NamedHDPublicKey, PSBT
from buidl.tx import Tx, TxIn, TxOut
from buidl.script import RedeemScript, address_to_script_pubkey


def _safe_get_child_hdpubkey(xfp_dict, xfp_hex, root_path, cnt):
    """
    Given an xfp_dict, inteligently traverse all the xpubs until you find one that can traverse to the given root_path
    """
    for base_path, xpub_obj in xfp_dict.get(xfp_hex, {}).items():
        child_path = get_unhardened_child_path(
            root_path=root_path,
            base_path=base_path,
        )
        if child_path:
            if base_path.count("/") != xpub_obj.depth:
                msg = f"xfp_hex {xfp_hex} for in/output #{cnt} base_path mismatch: {base_path} depth != {xpub_obj.depth} for {xpub_obj}"
                raise ValueError(msg)
            return xpub_obj.traverse(child_path)
    raise ValueError(
        f"xfp_hex {xfp_hex} with {root_path} for in/output #{cnt} not supplied in xpub_dict"
    )


def create_p2sh_multisig_psbt(
    public_key_records,
    input_dicts,
    output_dicts,
    fee_sats,
):
    """
    Helper method to create a p2sh multisig PSBT.

    network (testnet/mainnet) will be inferred from xpubs/tpubs.

    public_key_records are a list of entries that loom like this: [xfp_hex, xpub_b58, base_path]
    # TODO: turn this into a new object?
    """

    # initialize variables
    network = None
    tx_lookup, pubkey_lookup, redeem_lookup, hd_pubs = {}, {}, {}, {}

    # Use a nested default dict for increased readability
    # It's possible (though nonstandard) for one xfp to have multiple public_key_records in a multisig wallet
    # https://stackoverflow.com/a/19189356
    recursive_defaultdict = lambda: defaultdict(recursive_defaultdict)  # noqa: E731
    xfp_dict = recursive_defaultdict()

    # This at the child pubkey lookup that each input will traverse off of
    for xfp_hex, xpub_b58, base_path in public_key_records:
        hd_pubkey_obj = HDPublicKey.parse(xpub_b58)

        # We will use this dict/list structure for each input/ouput in the for-loops below
        xfp_dict[xfp_hex][base_path] = hd_pubkey_obj

        named_global_hd_pubkey_obj = NamedHDPublicKey.from_hd_pub(
            child_hd_pub=hd_pubkey_obj,
            xfp_hex=xfp_hex,
            # we're only going to base path level
            path=base_path,
        )
        hd_pubs[named_global_hd_pubkey_obj.serialize()] = named_global_hd_pubkey_obj

        if network is None:
            # Set the initial value
            network = hd_pubkey_obj.network
        else:
            # Confirm it hasn't changed
            if network != hd_pubkey_obj.network:
                raise MixedNetwork(
                    f"Mixed networks in public key records: {public_key_records}"
                )

    tx_ins = []
    for cnt, input_dict in enumerate(input_dicts):

        # Prev tx stuff
        prev_tx_dict = input_dict["prev_tx_dict"]
        prev_tx_obj = Tx.parse_hex(prev_tx_dict["hex"], network=network)
        tx_lookup[prev_tx_obj.hash()] = prev_tx_obj

        if prev_tx_dict["hash_hex"] != prev_tx_obj.hash().hex():
            raise ValueError(
                f"Hash digest mismatch for input #{cnt}: {prev_tx_dict['hash_hex']} != {prev_tx_obj.hash().hex()}"
            )

        # pubkey lookups needed for validation
        input_pubkey_hexes, total_input_sats = [], 0
        for xfp_hex, root_path in input_dict["path_dict"].items():
            # Get the correct xpub/path
            child_hd_pubkey = _safe_get_child_hdpubkey(
                xfp_dict=xfp_dict,
                xfp_hex=xfp_hex,
                root_path=root_path,
                cnt=cnt,
            )
            input_pubkey_hexes.append(child_hd_pubkey.sec().hex())

            # Enhance the PSBT
            named_hd_pubkey_obj = NamedHDPublicKey.from_hd_pub(
                child_hd_pub=child_hd_pubkey,
                xfp_hex=xfp_hex,
                path=root_path,
            )
            pubkey_lookup[named_hd_pubkey_obj.sec()] = named_hd_pubkey_obj

        utxo = prev_tx_obj.tx_outs[prev_tx_dict["output_idx"]]

        # Grab amount as developer safety check
        if prev_tx_dict["output_sats"] != utxo.amount:
            raise ValueError(
                f"Wrong number of sats for input #{cnt}! Expecting {prev_tx_dict['output_sats']} but got {utxo.amount}"
            )
        total_input_sats += utxo.amount

        redeem_script = RedeemScript.create_p2sh_multisig(
            quorum_m=input_dict["quorum_m"],
            pubkey_hexes=input_pubkey_hexes,
            sort_keys=True,  # TODO: allow legacy users to customize this?
            expected_addr=utxo.script_pubkey.address(network=network),
            expected_addr_network=network,
        )

        # Confirm address matches previous ouput
        if redeem_script.address(network=network) != utxo.script_pubkey.address(
            network=network
        ):
            raise ValueError(
                f"Invalid redeem script for input #{cnt}. Expecting {redeem_script.address(network=network)} but got {utxo.script_pubkey.address(network=network)}"
            )

        tx_in = TxIn(prev_tx=prev_tx_obj.hash(), prev_index=prev_tx_dict["output_idx"])
        tx_ins.append(tx_in)

        # For enhancing the PSBT for HWWs:
        redeem_lookup[redeem_script.hash160()] = redeem_script

    tx_outs = []
    for cnt, output_dict in enumerate(output_dicts):
        tx_out = TxOut(
            amount=output_dict["sats"],
            script_pubkey=address_to_script_pubkey(output_dict["address"]),
        )
        tx_outs.append(tx_out)

        if output_dict.get("path_dict"):
            # This output claims to be change, so we must validate it here
            output_pubkey_hexes = []
            for xfp_hex, root_path in output_dict["path_dict"].items():
                child_hd_pubkey = _safe_get_child_hdpubkey(
                    xfp_dict=xfp_dict,
                    xfp_hex=xfp_hex,
                    root_path=root_path,
                    cnt=cnt,
                )
                output_pubkey_hexes.append(child_hd_pubkey.sec().hex())

                # Enhance the PSBT
                named_hd_pubkey_obj = NamedHDPublicKey.from_hd_pub(
                    child_hd_pub=child_hd_pubkey,
                    xfp_hex=xfp_hex,
                    path=root_path,
                )
                pubkey_lookup[named_hd_pubkey_obj.sec()] = named_hd_pubkey_obj

            redeem_script = RedeemScript.create_p2sh_multisig(
                quorum_m=output_dict["quorum_m"],
                pubkey_hexes=output_pubkey_hexes,
                # We intentionally only allow change addresses to be lexicographically sorted
                sort_keys=True,
            )
            # Confirm address matches previous ouput
            if redeem_script.address(network=network) != output_dict["address"]:
                raise ValueError(
                    f"Invalid redeem script for output #{cnt}. Expecting {redeem_script.address(network=network)} but got {output_dict['address']}"
                )

            # For enhancing the PSBT for HWWs:
            redeem_lookup[redeem_script.hash160()] = redeem_script

    tx_obj = Tx(
        version=1,
        tx_ins=tx_ins,
        tx_outs=tx_outs,
        locktime=0,
        network=network,
        segwit=False,
    )

    # Safety check to try and prevent footgun

    calculated_fee_sats = total_input_sats - sum([tx_out.amount for tx_out in tx_outs])
    if fee_sats != calculated_fee_sats:
        raise ValueError(
            f"TX fee of {fee_sats} sats supplied != {calculated_fee_sats} sats calculated"
        )

    return PSBT.create(
        tx_obj=tx_obj,
        validate=True,
        tx_lookup=tx_lookup,
        pubkey_lookup=pubkey_lookup,
        redeem_lookup=redeem_lookup,
        witness_lookup=None,
        hd_pubs=hd_pubs,
    )



from buidl.helper import (
    bit_field_to_bytes,
    encode_varint,
    int_to_byte,
    int_to_little_endian,
    murmur3,
)
from buidl.network import GenericMessage


BIP37_CONSTANT = 0xFBA4C795


class BloomFilter:
    def __init__(self, size, function_count, tweak):
        self.size = size
        self.bit_field = [0] * (size * 8)
        self.function_count = function_count
        self.tweak = tweak

    def add(self, item):
        """Add an item to the filter"""
        # iterate self.function_count number of times
        for i in range(self.function_count):
            # BIP0037 spec seed is i*BIP37_CONSTANT + self.tweak
            seed = i * BIP37_CONSTANT + self.tweak
            # get the murmur3 hash given that seed
            h = murmur3(item, seed=seed)
            # set the bit at the hash mod the bitfield size (self.size*8)
            bit = h % (self.size * 8)
            # set the bit field at bit to be 1
            self.bit_field[bit] = 1

    def filter_bytes(self):
        return bit_field_to_bytes(self.bit_field)

    def filterload(self, flag=1):
        """Return a network message whose command is filterload"""
        # encode_varint self.size
        payload = encode_varint(self.size)
        # next is the self.filter_bytes()
        payload += self.filter_bytes()
        # function count is 4 bytes little endian
        payload += int_to_little_endian(self.function_count, 4)
        # tweak is 4 bytes little endian
        payload += int_to_little_endian(self.tweak, 4)
        # flag is 1 byte little endian
        payload += int_to_byte(flag)
        # return a GenericMessage with b'filterload' as the command
        return GenericMessage(b"filterload", payload)



#!/usr/bin/env python3
# coding: utf-8

import sys
from cmd import Cmd
from getpass import getpass
from platform import platform
from pkg_resources import DistributionNotFound, get_distribution

import buidl  # noqa: F401 (used below with pkg_resources for versioning)
from buidl.ecc import PrivateKey
from buidl.libsec_status import is_libsec_enabled
from buidl.psbt import PSBT


#####################################################################
# CLI UX
#####################################################################


# https://stackoverflow.com/questions/287871/how-to-print-colored-text-in-python
RESET_TERMINAL_COLOR = "\033[0m"


def blue_fg(string):
    return f"\033[34m{string}{RESET_TERMINAL_COLOR}"


def yellow_fg(string):
    return f"\033[93m{string}{RESET_TERMINAL_COLOR}"


def green_fg(string):
    return f"\033[32m{string}{RESET_TERMINAL_COLOR}"


def red_fg(string):
    return f"\033[31m{string}{RESET_TERMINAL_COLOR}"


def print_blue(string):
    print(blue_fg(string))


def print_yellow(string):
    print(yellow_fg(string))


def print_green(string):
    print(green_fg(string))


def print_red(string):
    print(red_fg(string))


def _get_buidl_version():
    try:
        return get_distribution("buidl").version
    except DistributionNotFound:
        return "Unknown"


def _get_bool(prompt, default=True):
    if default is True:
        yn = "[Y/n]"
    else:
        yn = "[y/N]"

    while True:
        response_str = input(blue_fg(f"{prompt} {yn}: ")).strip().lower()
        if response_str == "":
            return default
        if response_str in ("n", "no"):
            return False
        if response_str in ("y", "yes"):
            return True
        print_red("Please choose either y or n")


def _get_wif():
    prompt = blue_fg("Enter WIF (Wallet Import Format) to use for signing: ")
    while True:
        wif = getpass(prompt=prompt).strip()
        try:
            return PrivateKey.parse(wif)
        except Exception as e:
            print_red(f"Could not parse WIF: {e}")


def _get_psbt_obj(network):
    psbt_prompt = blue_fg(
        "Paste partially signed bitcoin transaction (PSBT) in base64 form: "
    )
    while True:
        psbt_b64 = input(psbt_prompt).strip()

        if not psbt_b64:
            continue

        try:
            psbt_obj = PSBT.parse_base64(psbt_b64, network=network)
        except Exception as e:
            print_red(f"Could not parse PSBT: {e}")
            continue

        # redundant but explicit
        if psbt_obj.validate() is not True:
            print_red("PSBT does not validate")
            continue

        return psbt_obj


#####################################################################
# Command Line App Code Starts Here
#####################################################################


class MyPrompt(Cmd):
    intro = (
        "Welcome to singlesweep, a stateless single sig sweeper that works with WIF and PSBTs.\n"
        "Single sig is DANGEROUS, this is an emergency recovery tool with NO WARRANTY OF ANY KIND.\n"
        "It is often used for collecting funds from old paper wallets.\n"
        "Type help or ? to list commands.\n"
    )
    prompt = "(₿) "  # the bitcoin symbol :)

    def __init__(self):
        super().__init__()

    def do_sweep(self, arg):
        """Sign a single-sig PSBT sweep transaction (1 output) using 1 WIF."""

        # We ask for this upfront so we can infer the network from it (PSBT doesn't have network info)
        # Users SHOULD only run this code on an airgap machine
        privkey_obj = _get_wif()
        network = privkey_obj.network

        # TODO: create a new helper method in pecc.py/cecc.py?
        expected_utxo_addr = privkey_obj.point.address(
            compressed=privkey_obj.compressed, network=network
        )
        print_yellow(
            f"Will attempt to spend from {network.upper()} {expected_utxo_addr}"
        )

        psbt_obj = _get_psbt_obj(network=network)
        tx_obj = psbt_obj.tx_obj

        try:
            psbt_described = psbt_obj.describe_p2pkh_sweep(privkey_obj=privkey_obj)
        except Exception as e:
            print_red(f"ABORTING WITHOUT SIGNING, could not describe PSBT:\n{e}")
            return

        # Gather TX info and validate
        print_yellow(psbt_described["tx_summary_text"])

        if _get_bool(prompt="In Depth Transaction View?", default=True):
            to_print = []
            to_print.append("DETAILED VIEW")
            to_print.append(f"Fee: {psbt_described['tx_fee_sats']:,} (unverified)")
            to_print.append(
                f"Total Input Sats Consumed: {psbt_described['total_input_sats']:,} (unverified)"
            )
            to_print.append(
                f"Total Output Sats Created: {psbt_described['output_spend_sats']:,}"
            )
            to_print.append(f"Lock Time: {psbt_described['locktime']:,}")
            to_print.append(
                f"RBF: {'Enabled' if psbt_described['is_rbf_able'] else 'DISABLED'}"
            )
            to_print.append(
                f"Size: {psbt_described['tx_size_bytes']} bytes (will increase after signing)"
            )
            to_print.append("-" * 80)
            to_print.append(f"{len(psbt_described['inputs_desc'])} Input(s):")
            for cnt, input_desc in enumerate(psbt_described["inputs_desc"]):
                to_print.append(f"  Input #{cnt}")
                for k, v in input_desc.items():
                    if k == "sats":
                        # Comma separate ints
                        val = f"{v:,} (unverified)"
                    else:
                        val = v
                    to_print.append(f"    {k}: {val}")
            to_print.append("-" * 80)
            to_print.append(f"{len(psbt_described['outputs_desc'])} Output(s):")
            for cnt, output_desc in enumerate(psbt_described["outputs_desc"]):
                to_print.append(f"  Output #{cnt}")
                for k, v in output_desc.items():
                    if k == "sats":
                        # Comma separate ints
                        val = f"{v:,}"
                    else:
                        val = v
                    to_print.append(f"    {k}: {val}")
            print_yellow("\n".join(to_print))

        if not _get_bool(prompt="Sign this transaction?", default=True):
            print_yellow(f"Transaction {tx_obj.id()} NOT signed")
            return

        # Sign the TX
        # TODO: would prefer to use psbt_obj.sign_with_private_keys(), but that requires NamedPublicKeys that we don't have (no paths in PSBT)
        for cnt, _ in enumerate(tx_obj.tx_ins):
            was_signed = tx_obj.sign_p2pkh(input_index=cnt, private_key=privkey_obj)
            if was_signed is not True:
                print_red("PSBT was NOT signed")

        print_yellow(f"SIGNED TX {tx_obj.hash().hex()} has the following hex:\n")
        print_green(tx_obj.serialize().hex())
        print_yellow("\nYou can be broadcast this hex via:")
        print_yellow(" - Your bitcoin core node")
        print_yellow(
            ' - "pushtx" block explorers (Blockstream, BlockCypher, Blockchain.com, etc), mining pools, Electrum SPV network, etc'
        )
        print_yellow(
            ' - Electrum signing of a previously unsigned transaction: "Combine" > "Merge Signatures From"\n'
        )

    def do_version_info(self, arg):
        """Print program settings for debug purposes"""

        to_print = [
            f"buidl Version: {_get_buidl_version()}",
            f"Python Version: {sys.version_info}",
            f"Platform: {platform()}",
            f"libsecp256k1 Configured: {is_libsec_enabled()}",
        ]
        print_yellow("\n".join(to_print))

    def do_exit(self, arg):
        """Exit Program"""
        print_yellow("\nNo data saved")
        return True


if __name__ == "__main__":
    try:
        MyPrompt().cmdloop()
    except KeyboardInterrupt:
        print_yellow("\nNo data saved")



from unittest import TestCase

from buidl.descriptor import (
    calc_core_checksum,
    is_valid_xfp_hex,
    P2WSHSortedMulti,
    parse_full_key_record,
    parse_partial_key_record,
)
from buidl.hd import HDPrivateKey


class P2WSHMultiTest(TestCase):
    def test_p2wsh_1of4(self):
        valid_output_record = """wsh(sortedmulti(1,[c7d0648a/48h/1h/0h/2h]tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr/0/*,[12980eed/48h/1h/0h/2h]tpubDEkXGoQhYLFnYyzUGadtceUKbzVfXVorJEdo7c6VKJLHrULhpSVLC7fo89DDhjHmPvvNyrun2LTWH6FYmHh5VaQYPLEqLviVQKh45ufz8Ae/0/*,[3a52b5cd/48h/1h/0h/2h]tpubDFdbVee2Zna6eL9TkYBZDJVJ3RxGYWgChksXBRgw6y6PU1jWPTXUqag3CBMd6VDwok1hn5HZGvg6ujsTLXykrS3DwbxqCzEvWoT49gRJy7s/0/*,[f7d04090/48h/1h/0h/2h]tpubDF7FTuPECTePubPXNK73TYCzV3nRWaJnRwTXD28kh6Fz4LcaRzWwNtX153J7WeJFcQB2T6k9THd424Kmjs8Ps1FC1Xb81TXTxxbGZrLqQNp/0/*))#tatkmj5q"""

        p2wsh_sortedmulti_obj = P2WSHSortedMulti.parse(valid_output_record)

        # Should recreate itself
        self.assertEqual(str(p2wsh_sortedmulti_obj), valid_output_record)

        self.assertEqual(p2wsh_sortedmulti_obj.network, "testnet")
        self.assertEqual(p2wsh_sortedmulti_obj.quorum_m, 1)

        self.assertEqual(p2wsh_sortedmulti_obj.checksum, "tatkmj5q")

        expected_change_addrs = [
            "tb1qf454te8pvz4txevejg8s8tx5kkyfxgtkpg6tu5xphnyf6l2gcjss5zw0jx",
            "tb1q0lh5an3hep9s57c5xkpyv0yldy825kzzwt888u0qfnu3nqkndrqqjuajuj",
            "tb1qcklg00ymx85x7f5vzll3zypd2epywdxmhx05k7395e470pth8g8qvh62kw",
        ]
        expected_receive_addrs = [
            "tb1qlrjv2ek09g9aplga83j9mfvelnt6qymen9gd49kpezdz2g5pgwnsfmrucp",
            "tb1qn2xhgxqxqcs8cl36f7efgg7jvreus4x6959hnc6mfmygnz435dksa39ygr",
            "tb1q2lzh628dmylpf9gr869lgyq9fcc9xqat7unpumnmmn5nph6447cs40k7mw",
        ]

        for cnt, change_addr in enumerate(expected_change_addrs):
            self.assertEqual(
                change_addr,
                p2wsh_sortedmulti_obj.get_address(is_change=True, offset=cnt),
            )

        for cnt, receive_addr in enumerate(expected_receive_addrs):
            self.assertEqual(
                receive_addr,
                p2wsh_sortedmulti_obj.get_address(is_change=False, offset=cnt),
            )

        expected_key_records = [
            {
                "xfp": "c7d0648a",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr",
                "account_index": 0,
            },
            {
                "xfp": "12980eed",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDEkXGoQhYLFnYyzUGadtceUKbzVfXVorJEdo7c6VKJLHrULhpSVLC7fo89DDhjHmPvvNyrun2LTWH6FYmHh5VaQYPLEqLviVQKh45ufz8Ae",
                "account_index": 0,
            },
            {
                "xfp": "3a52b5cd",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDFdbVee2Zna6eL9TkYBZDJVJ3RxGYWgChksXBRgw6y6PU1jWPTXUqag3CBMd6VDwok1hn5HZGvg6ujsTLXykrS3DwbxqCzEvWoT49gRJy7s",
                "account_index": 0,
            },
            {
                "xfp": "f7d04090",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDF7FTuPECTePubPXNK73TYCzV3nRWaJnRwTXD28kh6Fz4LcaRzWwNtX153J7WeJFcQB2T6k9THd424Kmjs8Ps1FC1Xb81TXTxxbGZrLqQNp",
                "account_index": 0,
            },
        ]
        self.assertEqual(p2wsh_sortedmulti_obj.key_records, expected_key_records)

        # Parse again, but without the checksum (should regenerate it)
        valid_output_record_sans_checksum = valid_output_record.split("#")[0]
        p2wsh_sortedmulti_obj_regenerated = P2WSHSortedMulti.parse(
            valid_output_record_sans_checksum
        )
        self.assertEqual(
            p2wsh_sortedmulti_obj_regenerated.checksum, p2wsh_sortedmulti_obj.checksum
        )
        # Redundant
        self.assertEqual(
            str(p2wsh_sortedmulti_obj_regenerated), str(p2wsh_sortedmulti_obj)
        )

        # Parse again, but with an invalid checksum, which should throw an error
        invalid_checksums = ["", "a", "a" * 9, "X"]
        for invalid_checksum in invalid_checksums:
            with self.assertRaises(ValueError):
                P2WSHSortedMulti.parse(
                    valid_output_record_sans_checksum + "#" + invalid_checksum
                )

        # Parse again, but with the wrong (but valid) checksum, which should throw an error
        with self.assertRaises(ValueError):
            P2WSHSortedMulti.parse(valid_output_record_sans_checksum + "#" + "a" * 8)

        # manually checked on 2021-06-08 that when imported to Caravan it generates the same addresses as buidl
        want = """{"name": "p2wsh", "addressType": "P2WSH", "network": "testnet", "client": {"type": "public"}, "quorum": {"requiredSigners": 1, "totalSigners": 4}, "extendedPublicKeys": [{"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr", "xfp": "c7d0648a", "name": "Seed A"}, {"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDEkXGoQhYLFnYyzUGadtceUKbzVfXVorJEdo7c6VKJLHrULhpSVLC7fo89DDhjHmPvvNyrun2LTWH6FYmHh5VaQYPLEqLviVQKh45ufz8Ae", "xfp": "12980eed", "name": "Seed B"}, {"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDFdbVee2Zna6eL9TkYBZDJVJ3RxGYWgChksXBRgw6y6PU1jWPTXUqag3CBMd6VDwok1hn5HZGvg6ujsTLXykrS3DwbxqCzEvWoT49gRJy7s", "xfp": "3a52b5cd", "name": "Seed C"}, {"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDF7FTuPECTePubPXNK73TYCzV3nRWaJnRwTXD28kh6Fz4LcaRzWwNtX153J7WeJFcQB2T6k9THd424Kmjs8Ps1FC1Xb81TXTxxbGZrLqQNp", "xfp": "f7d04090", "name": "Seed D"}], "startingAddressIndex": 0}"""
        self.assertEqual(p2wsh_sortedmulti_obj.caravan_export(), want)

    def test_p2wsh_2of3(self):
        valid_output_record = "wsh(sortedmulti(2,[c7d0648a/48h/1h/0h/2h]tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr/0/*,[12980eed/48h/1h/0h/2h]tpubDEkXGoQhYLFnYyzUGadtceUKbzVfXVorJEdo7c6VKJLHrULhpSVLC7fo89DDhjHmPvvNyrun2LTWH6FYmHh5VaQYPLEqLviVQKh45ufz8Ae/0/*,[f7d04090/48h/1h/0h/2h]tpubDF7FTuPECTePubPXNK73TYCzV3nRWaJnRwTXD28kh6Fz4LcaRzWwNtX153J7WeJFcQB2T6k9THd424Kmjs8Ps1FC1Xb81TXTxxbGZrLqQNp/0/*))#0stzl64e"
        p2wsh_sortedmulti_obj = P2WSHSortedMulti.parse(valid_output_record)
        # Should recreate itself
        self.assertEqual(str(p2wsh_sortedmulti_obj), valid_output_record)

        self.assertEqual(p2wsh_sortedmulti_obj.network, "testnet")
        self.assertEqual(p2wsh_sortedmulti_obj.checksum, "0stzl64e")
        self.assertEqual(
            p2wsh_sortedmulti_obj.get_address(),
            "tb1q0cy5x39ezyvc4pfydrqedng0h9arh2hcw8lpfa6e9ama7ky7cffsmzmgx8",
        )

        expected_key_records = [
            {
                "xfp": "c7d0648a",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr",
                "account_index": 0,
            },
            {
                "xfp": "12980eed",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDEkXGoQhYLFnYyzUGadtceUKbzVfXVorJEdo7c6VKJLHrULhpSVLC7fo89DDhjHmPvvNyrun2LTWH6FYmHh5VaQYPLEqLviVQKh45ufz8Ae",
                "account_index": 0,
            },
            {
                "xfp": "f7d04090",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDF7FTuPECTePubPXNK73TYCzV3nRWaJnRwTXD28kh6Fz4LcaRzWwNtX153J7WeJFcQB2T6k9THd424Kmjs8Ps1FC1Xb81TXTxxbGZrLqQNp",
                "account_index": 0,
            },
        ]
        self.assertEqual(p2wsh_sortedmulti_obj.key_records, expected_key_records)

        # manually checked on 2021-06-08 that when imported to Caravan it generates the same addresses as buidl
        want = """{"name": "p2wsh", "addressType": "P2WSH", "network": "testnet", "client": {"type": "public"}, "quorum": {"requiredSigners": 2, "totalSigners": 3}, "extendedPublicKeys": [{"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr", "xfp": "c7d0648a", "name": "Seed A"}, {"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDEkXGoQhYLFnYyzUGadtceUKbzVfXVorJEdo7c6VKJLHrULhpSVLC7fo89DDhjHmPvvNyrun2LTWH6FYmHh5VaQYPLEqLviVQKh45ufz8Ae", "xfp": "12980eed", "name": "Seed B"}, {"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDF7FTuPECTePubPXNK73TYCzV3nRWaJnRwTXD28kh6Fz4LcaRzWwNtX153J7WeJFcQB2T6k9THd424Kmjs8Ps1FC1Xb81TXTxxbGZrLqQNp", "xfp": "f7d04090", "name": "Seed C"}], "startingAddressIndex": 0}"""
        self.assertEqual(p2wsh_sortedmulti_obj.caravan_export(), want)

    def test_p2wsh_1of2_sorted_and_unsorted(self):
        TESTNET_DEFAULT_PATH = "m/48h/1h/0h/2h"
        # https://github.com/satoshilabs/slips/blob/master/slip-0132.md
        TESTNET_VERSION_BYTE = bytes.fromhex("043587cf")

        key_records = []
        for mnenmonic in ("invest ", "sell "):
            hd_priv_obj = HDPrivateKey.from_mnemonic(mnenmonic * 12)
            xfp_hex = hd_priv_obj.fingerprint().hex()
            xpub = hd_priv_obj.traverse(path=TESTNET_DEFAULT_PATH).xpub(
                TESTNET_VERSION_BYTE
            )
            key_records.append(
                {
                    "xfp": xfp_hex,
                    "path": TESTNET_DEFAULT_PATH,
                    "xpub_parent": xpub,
                    "account_index": 0,
                }
            )

        p2wsh_sorted_obj = P2WSHSortedMulti(quorum_m=1, key_records=key_records)
        expected = "wsh(sortedmulti(1,[aa917e75/48h/1h/0h/2h]tpubDEZRP2dRKoGRJnR9zn6EoLouYKbYyjFsxywgG7wMQwCDVkwNvoLhcX1rTQipYajmTAF82kJoKDiNCgD4wUPahACE7n1trMSm7QS8B3S1fdy/0/*,[2553c4b8/48h/1h/0h/2h]tpubDEiNuxUt4pKjKk7khdv9jfcS92R1WQD6Z3dwjyMFrYj2iMrYbk3xB5kjg6kL4P8SoWsQHpd378RCTrM7fsw4chnJKhE2kfbfc4BCPkVh6g9/0/*))#t0v98kwu"
        self.assertEqual(expected, str(p2wsh_sorted_obj))
        self.assertEqual(
            p2wsh_sorted_obj.get_address(1),
            "tb1q6y5dh62l40q9de8k53sjekqz2zcn9dfs55g5v8pjmd7ehg08gk0sms9q4l",
        )

        # Test unsorted

        # These sometimes produce the same address (should be 50% on average)
        addr_matches = [False, True, True, True, True, False, True, False, True, False]
        for cnt, addr_match in enumerate(addr_matches):
            sorted_addr = p2wsh_sorted_obj.get_address(cnt, sort_keys=True)
            unsorted_addr = p2wsh_sorted_obj.get_address(cnt, sort_keys=False)
            if addr_match:
                self.assertEqual(sorted_addr, unsorted_addr)
            else:
                self.assertNotEqual(sorted_addr, unsorted_addr)

        # manually checked on 2021-06-08 that when imported to Caravan it generates the same addresses as buidl
        want = """{"name": "foo", "addressType": "P2WSH", "network": "testnet", "client": {"type": "public"}, "quorum": {"requiredSigners": 1, "totalSigners": 2}, "extendedPublicKeys": [{"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDEZRP2dRKoGRJnR9zn6EoLouYKbYyjFsxywgG7wMQwCDVkwNvoLhcX1rTQipYajmTAF82kJoKDiNCgD4wUPahACE7n1trMSm7QS8B3S1fdy", "xfp": "aa917e75", "name": "alice"}, {"bip32Path": "m/48'/1'/0'/2'", "xpub": "tpubDEiNuxUt4pKjKk7khdv9jfcS92R1WQD6Z3dwjyMFrYj2iMrYbk3xB5kjg6kL4P8SoWsQHpd378RCTrM7fsw4chnJKhE2kfbfc4BCPkVh6g9", "xfp": "2553c4b8", "name": "bob"}], "startingAddressIndex": 0}"""
        self.assertEqual(
            p2wsh_sorted_obj.caravan_export(
                wallet_name="foo", key_record_names=["alice", "bob"]
            ),
            want,
        )

    def test_mixed_slip132_p2wsh_sortedmulti(self):
        # this should coalesce the values into non-slip132 (p2wsh sortedmulti already conveys the script type unambiguously)
        quorum_m = 1
        key_records = [
            {
                # investx12:
                "xfp": "aa917e75",
                "path": "m/48h/1h/0h/2h",
                "xpub_parent": "tpubDEZRP2dRKoGRJnR9zn6EoLouYKbYyjFsxywgG7wMQwCDVkwNvoLhcX1rTQipYajmTAF82kJoKDiNCgD4wUPahACE7n1trMSm7QS8B3S1fdy",
                "account_index": 0,
            },
            {
                # sellx12:
                "xfp": "2553c4b8",
                "path": "m/48h/1h/0h/2h/2046266013/1945465733/1801020214/1402692941",
                "xpub_parent": "Vpub5uMrp2GYpnHN8BkjvXpP71TuZ8BDqu61PPcwEKSzE9Mcuow727mUJNsDsKdzAiupHXea5F7ZxD9SaSQvbr1hvpNjrijJQ2J46VQjc5yEcm8",
                "account_index": 0,
            },
        ]
        p2wsh_sortedmulti_obj = P2WSHSortedMulti(
            quorum_m, key_records, sort_key_records=False
        )
        # notice that both values are tpub, despite ingesting Vpub
        want = "wsh(sortedmulti(1,[aa917e75/48h/1h/0h/2h]tpubDEZRP2dRKoGRJnR9zn6EoLouYKbYyjFsxywgG7wMQwCDVkwNvoLhcX1rTQipYajmTAF82kJoKDiNCgD4wUPahACE7n1trMSm7QS8B3S1fdy/0/*,[2553c4b8/48h/1h/0h/2h/2046266013/1945465733/1801020214/1402692941]tpubDNVvpMhdGTmQg1AT6muju2eUWPXWWAtUSyc1EQ2MxJ2s97fMqFZQbpzQM4gU8bwzfFM7KBpSXRJ5v2Wu8sY2GF5ZpXm3qy8GLArZZNM1Wru/0/*))#0lfdttke"
        self.assertEqual(str(p2wsh_sortedmulti_obj), want)

    def test_invalid_p2wsh_sortedmulti(self):
        # Notice the mix of xpub and tpub
        output_record = """wsh(sortedmulti(1,[c7d0648a/48h/1h/0h/2h]tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr/0/*,[12980eed/48h/1h/0h/2h]xpub6ENtkZb1q4JLHBocpPeoQj8xGsQ1Y7Jnkc3Vm43LyPaQ7BfzkDeF3fzxt78SBELXc2PUNHPuVEdTaukwNRqqc8xFKjVXfQ4FpN6eKqe6y9E/0/*))#tatkmj5q"""

        with self.assertRaises(ValueError) as fail:
            print("fail")
            P2WSHSortedMulti.parse(output_record)

        self.assertIn(
            "Network mismatch: network is set to testnet but ", str(fail.exception)
        )


class ParseTest(TestCase):
    def test_valid_key_record_regular(self):
        key_record = "[c7d0648a/48h/1h/0h/2h]tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr/0/*"
        results = parse_full_key_record(key_record)
        want = {
            "xfp": "c7d0648a",
            "path": "m/48h/1h/0h/2h",
            "xpub_parent": "tpubDEpefcgzY6ZyEV2uF4xcW2z8bZ3DNeWx9h2BcwcX973BHrmkQxJhpAXoSWZeHkmkiTtnUjfERsTDTVCcifW6po3PFR1JRjUUTJHvPpDqJhr",
            "xpub_child": "tpubDHXhgZEb9KfoFAuPQ5X6nayFrgifHEb3EUAbbs3EvwboxjttP4ekmPPz4NPRDE7p3q87DQH2TbNyxUmYGf2GNiSTfXj4Q5CfVgrpZuDEsak",
            "account_index": 0,
            "network": "testnet",
        }
        self.assertEqual(results, want)

    def test_valid_key_record_slip132(self):
        # oil * 12
        key_record = "[2a77e0a6/48h/1h/0h/2h]Vpub5mvQbnmqKfpPjWfAZEw5Xjdr6UjnjyZEirzrhNMSuKjL8Qfd3nqLBkrBrVXNeMgKCjPXbyLnSCn6qcD8fHQCkNnNLnkpQtY3sh4MHmywvbe"
        results = parse_partial_key_record(key_record)
        want = {
            "xfp": "2a77e0a6",
            "path": "m/48h/1h/0h/2h",
            "xpub": "Vpub5mvQbnmqKfpPjWfAZEw5Xjdr6UjnjyZEirzrhNMSuKjL8Qfd3nqLBkrBrVXNeMgKCjPXbyLnSCn6qcD8fHQCkNnNLnkpQtY3sh4MHmywvbe",
            "network": "testnet",
        }
        self.assertEqual(results, want)


class CoreChecksumTest(TestCase):
    def test_sd_descriptor_checksum_a(self):
        # https://github.com/cryptoadvance/specter-desktop/blob/c1ce0e982a0552883cc339ecf5ee6860574e79b7/tests/test_util_descriptor.py
        descriptor_ex_checksum = "sh(multi(2,[00000000/111'/222]xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4koxb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL,xpub68NZiKmJWnxxS6aaHmn81bvJeTESw724CRDs6HbuccFQN9Ku14VQrADWgqbhhTHBaohPX4CjNLf9fq9MYo6oDaPPLPxSb7gwQN3ih19Zm4Y/0))"
        self.assertEqual(calc_core_checksum(descriptor_ex_checksum), "tjg09x5t")

    def test_sd_descriptor_checksum_b(self):
        # https://github.com/cryptoadvance/specter-desktop/blob/c1ce0e982a0552883cc339ecf5ee6860574e79b7/tests/test_util_descriptor.py
        descriptor_ex_checksum = "sh(wsh(sortedmulti(2,xpub6DkFAXWQ2dHxnMKoSBogHrw1rgNJKR4umdbnNVNTYeCGcduxWnNUHgGptqEQWPKRmeW4Zn4FHSbLMBKEWYaMDYu47Ytg6DdFnPNt8hwn5mE/1,xpub6DiXipxEgSYqTw3xX2apub7vzsC5gBzmikxriTRnfKRKQjUSpiGQ9XzyFktkVLTGVGF5emH8up1qtsyw726rvnmzHRU8cHH8gDxeLMXSkYE/1,xpub6FHZCoNb3tg3mxjcXsQx1xLpNmod6woECf2fB4nQbe9NXbvha2ucpDpnGbTFF68KUMUr1hNQ9E5jVEvpT2kUkVmFVDrJawcbgXzDpJc2hkF/2)))"
        self.assertEqual(calc_core_checksum(descriptor_ex_checksum), "mgjhd0rk")

    def test_sd_descriptor_checksum_c(self):
        # https://github.com/cryptoadvance/specter-desktop/blob/c1ce0e982a0552883cc339ecf5ee6860574e79b7/tests/test_util_descriptor.py
        descriptor_ex_checksum = "sh(wsh(sortedmulti(2,029dfee2aaa23e2220476c34eda9a76591c1257f8dfce54e42ff014f922ede0838,03151d5b21c6491915e7a103bff913b4d85246c8209a342bb7104850e4cb394686,03646d8e624fedb63739e7963d0c7ad368a7f7935557b2b28c4c954882b19fe6e1)))"
        self.assertEqual(calc_core_checksum(descriptor_ex_checksum), "rzmdthwy")


class SeedFingerprintTest(TestCase):
    def test_valid_xfps(self):
        self.assertTrue(is_valid_xfp_hex("a" * 8))
        self.assertTrue(is_valid_xfp_hex("abcdef01"))
        self.assertTrue(is_valid_xfp_hex("01234567"))
        self.assertTrue(is_valid_xfp_hex("deadbeef"))

    def test_invalid_xfps(self):
        invalid_xfps = [
            "a" * 7,
            "a" * 9,
            "hello123",
        ]
        for invalid in invalid_xfps:
            self.assertFalse(is_valid_xfp_hex(invalid))
